#!/usr/bin/env python
# crate_anon/crateweb/core/utils.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import datetime
import logging
import urllib.parse
from typing import Any, Dict, List, Union

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.paginator import Paginator, EmptyPage, Page, PageNotAnInteger
from django.db.models import QuerySet
from django.http import QueryDict
from django.http.request import HttpRequest
from django.utils import timezone
from crate_anon.crateweb.userprofile.models import get_per_page

log = logging.getLogger(__name__)


# =============================================================================
# User tests/user profile
# =============================================================================

def is_superuser(user: settings.AUTH_USER_MODEL) -> bool:
    """
    Function for user with decorator, e.g.
        @user_passes_test(is_superuser)
        
    Superuser equates to Research Database Manager.
    """
    # https://docs.djangoproject.com/en/dev/topics/auth/default/#django.contrib.auth.decorators.user_passes_test  # noqa
    return user.is_superuser


def is_developer(user: settings.AUTH_USER_MODEL) -> bool:
    if not user.is_authenticated():
        return False  # won't have a profile
    return user.profile.is_developer


def is_clinician(user: settings.AUTH_USER_MODEL) -> bool:
    if not user.is_authenticated():
        return False  # won't have a profile
    return user.profile.is_clinician


# =============================================================================
# Forms
# =============================================================================

def paginate(request: HttpRequest,
             all_items: Union[QuerySet, List[Any]],
             per_page: int = None) -> Page:
    if per_page is None:
        per_page = get_per_page(request)
    paginator = Paginator(all_items, per_page)
    requested_page = request.GET.get('page')
    try:
        return paginator.page(requested_page)
    except PageNotAnInteger:
        return paginator.page(1)
    except EmptyPage:
        return paginator.page(paginator.num_pages)


# =============================================================================
# URL creation
# =============================================================================

def url_with_querystring(path: str,
                         querydict: Dict[str, Any] = None,
                         **kwargs) -> str:
    """Add GET arguments to a URL from named arguments or a QueryDict."""
    if querydict is not None and not isinstance(querydict, QueryDict):
        raise ValueError("Bad querydict value")
    if querydict and kwargs:
        raise ValueError("Pass either a QueryDict or kwargs")
    if querydict:
        querystring = querydict.urlencode()
    else:
        querystring = urllib.parse.urlencode(kwargs)
    return path + '?' + querystring


def site_absolute_url(path: str) -> str:
    """
    Returns an absolute URL for the site, given a relative part.
    Use like:
        url = site_absolute_url(static('red.png'))
            ... determined in part by STATIC_URL.
        url = site_absolute_url(reverse('clinician_response', args=[self.id]))
            ... determined by SCRIPT_NAME or FORCE_SCRIPT_NAME
            ... which is context-dependent: see below

    We need to generate links to our site outside the request environment, e.g.
    for inclusion in e-mails, even when we're generating the e-mails offline
    via Celery. There's no easy way to do this automatically (site path
    information comes in only via requests), so we put it in the settings.

    See also:
        http://stackoverflow.com/questions/4150258/django-obtaining-the-absolute-url-without-access-to-a-request-object  # noqa
        https://fragmentsofcode.wordpress.com/2009/02/24/django-fully-qualified-url/  # noqa

    ---------------------------------------------------------------------------
    IMPORTANT
    ---------------------------------------------------------------------------
    BEWARE: reverse() will produce something different inside a request and
    outside it.
        http://stackoverflow.com/questions/32340806/django-reverse-returns-different-values-when-called-from-wsgi-or-shell  # noqa

    So the only moderately clean way of doing this is to do this in the Celery
    backend jobs, for anything that uses Django URLs (e.g. reverse) -- NOT
    necessary for anything using only static URLs (e.g. pictures in PDFs).

        from django.conf import settings
        from django.core.urlresolvers import set_script_prefix

        set_script_prefix(settings.FORCE_SCRIPT_NAME)

    But that does at least mean we can use the same method for static and
    Django URLs.
    """
    url = settings.DJANGO_SITE_ROOT_ABSOLUTE_URL + path
    log.debug("site_absolute_url: {} -> {}".format(path, url))
    return url


# =============================================================================
# Formatting
# =============================================================================

def get_friendly_date(date: datetime.datetime) -> str:
    if date is None:
        return ""
    try:
        return date.strftime("%d %B %Y")  # e.g. 03 December 2013
    except Exception as e:
        raise type(e)(str(e) + ' [value was {}]'.format(repr(date)))


def modelrepr(instance) -> str:
    """Default repr version of a Django model object, for debugging."""
    elements = []
    # noinspection PyProtectedMember
    for fieldname in [f.name for f in instance._meta.get_fields()]:
        try:
            value = repr(getattr(instance, fieldname))
        except ObjectDoesNotExist:
            value = "<RelatedObjectDoesNotExist>"
        elements.append("{}={}".format(fieldname, value))
    return "<{} <{}>>".format(type(instance).__name__,
                              ", ".join(elements))
    # - type(instance).__name__ gives the Python class name from an instance
    # - ... as does ModelClass.__name__ but we don't have that directly here
    # - instance._meta.model_name gives a lower-case version


# =============================================================================
# Date/time
# =============================================================================

def string_time_now() -> str:
    """Returns current time in short-form ISO-8601 UTC, for filenames."""
    return timezone.now().strftime("%Y%m%dT%H%M%SZ")



#!/usr/bin/env python
# crate_anon/common/lang.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
import sys
import traceback
from typing import Any, Dict, Iterable, List, Optional, Tuple

log = logging.getLogger(__name__)


# =============================================================================
# Dictionary handling
# =============================================================================

def get_case_insensitive_dict_key(d: Dict, k: str) -> Optional[str]:
    for key in d.keys():
        if k.lower() == key.lower():
            return key
    return None


def rename_kwarg(kwargs: Dict[str, Any], old: str, new: str) -> None:
    kwargs[new] = kwargs.pop(old)


def merge_two_dicts(x: Dict, y: Dict) -> Dict:
    """
    Given two dicts, merge them into a new dict as a shallow copy, e.g.
        z = merge_two_dicts(x, y)
    If you can guarantee Python 3.5, then a simpler syntax is:
        z = {**x, **y}
    See http://stackoverflow.com/questions/38987
    """
    z = x.copy()
    z.update(y)
    return z


def rename_keys_in_dict(d: Dict[str, Any], renames: Dict[str, str]) -> None:
    """
    Renames, in place, the keys in d according to the mapping in "renames".
    """
    # https://stackoverflow.com/questions/4406501/change-the-name-of-a-key-in-dictionary  # noqa
    for old_key, new_key in renames.items():
        if old_key in d:
            if new_key in d:
                raise ValueError(
                    "rename_keys_in_dict: renaming {} -> {} but new key "
                    "already exists".format(repr(old_key), repr(new_key)))
            d[new_key] = d.pop(old_key)


# =============================================================================
# List handling
# =============================================================================

def unique_list(seq: Iterable[Any]) -> List[Any]:
    # http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-whilst-preserving-order  # noqa
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


# =============================================================================
# Exception handling
# =============================================================================

def add_info_to_exception(err: Exception, info: Dict) -> None:
    # http://stackoverflow.com/questions/9157210/how-do-i-raise-the-same-exception-with-a-custom-message-in-python  # noqa
    if not err.args:
        err.args = ('', )
    err.args += (info, )


def recover_info_from_exception(err: Exception) -> Dict:
    if len(err.args) < 1:
        return {}
    info = err.args[-1]
    if not isinstance(info, dict):
        return {}
    return info


def die(exc: Exception = None, exit_code: int = 1) -> None:
    """
    It is not clear that Python guarantees to exit with a non-zero exit code
    (errorlevel in DOS/Windows) upon an unhandled exception. So this function
    produces the usual stack trace then dies.

    http://stackoverflow.com/questions/9555133/e-printstacktrace-equivalent-in-python  # noqa

    Test code:

import logging
import sys
import traceback
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger()

def fail():
    try:
        x = 1/0
    except Exception as exc:
        die(exc)


    Then call
        fail()
    ... which should exit Python; then from Linux:
        echo $?  # show exit code

    """
    if exc:
        lines = traceback.format_exception(
            None,  # etype: ignored
            exc,
            exc.__traceback__)  # https://www.python.org/dev/peps/pep-3134/
        msg = "".join(lines)
        # Method 1:
        # print("".join(lines), file=sys.stderr, flush=True)
        # Method 2:
        log.critical(msg)
    log.critical("Exiting with exit code {}".format(exit_code))
    sys.exit(exit_code)


# =============================================================================
# __repr__ aids
# =============================================================================
# The repr() function often attempts to return something suitable for eval();
# failing that, it usually shows an address.
# https://docs.python.org/3/library/functions.html#repr

def _repr_result(obj: Any, elements: List[str],
                 with_addr: bool = False) -> str:
    if with_addr:
        return "<{qualname}({elements}) at {addr}>".format(
            qualname=obj.__class__.__qualname__,
            elements=", ".join(elements),
            addr=hex(id(obj)),
        )
    else:
        return "{qualname}({elements})".format(
            qualname=obj.__class__.__qualname__,
            elements=", ".join(elements),
        )


def auto_repr(obj: Any, with_addr: bool = False) -> str:
    """
    Convenience function for repr().
    Works its way through the object's __dict__ and reports accordingly.
    """
    elements = ["{}={}".format(k, repr(v)) for k, v in obj.__dict__.items()]
    return _repr_result(obj, elements, with_addr=with_addr)


def simple_repr(obj: Any, attrnames: List[str],
                with_addr: bool = False) -> str:
    """
    Convenience function for repr().
    Works its way through a list of attribute names, and creates a repr()
    assuming that parameters to the constructor have the same names.
    """
    elements = ["{}={}".format(name, repr(getattr(obj, name)))
                for name in attrnames]
    return _repr_result(obj, elements, with_addr=with_addr)


def mapped_repr(obj: Any, attributes: List[Tuple[str, str]],
                with_addr: bool = False) -> str:
    """
    Convenience function for repr().
    Takes a list of tuples: (attr_name, init_param_name).
    """
    elements = ["{}={}".format(init_param_name, repr(getattr(obj, attr_name)))
                for attr_name, init_param_name in attributes]
    return _repr_result(obj, elements, with_addr=with_addr)


def mapped_repr_stripping_underscores(obj: Any, attrnames: List[str],
                                      with_addr: bool = False) -> str:
    """
    Convenience function for repr().
    Here, you pass a list of internal attributes, and it assumes that the
    __init__() parameter names have the leading underscore dropped.
    """
    attributes = []
    for attr_name in attrnames:
        if attr_name.startswith('_'):
            init_param_name = attr_name[1:]
        else:
            init_param_name = attr_name
        attributes.append((attr_name, init_param_name))
    return mapped_repr(obj, attributes, with_addr=with_addr)



#!/usr/bin/env python
# crate_anon/crateweb/consent/utils.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import datetime
# from functools import lru_cache
import os
from typing import Any, Dict, Optional

from django.conf import settings
from django.core.exceptions import ValidationError
from django.template.loader import render_to_string

from crate_anon.crateweb.extra.django_cache_fn import django_cache_function


def read_static_file_contents(filename: str) -> str:
    with open(os.path.join(settings.LOCAL_STATIC_DIR, filename)) as f:
        return f.read()


def pdf_css(patient: bool = True) -> str:
    contents = read_static_file_contents('base.css')
    context = {
        'fontsize': (settings.PATIENT_FONTSIZE
                     if patient else settings.RESEARCHER_FONTSIZE),
    }
    contents += render_to_string('pdf.css', context)
    return contents


@django_cache_function(timeout=None)
# @lru_cache(maxsize=None)
def pdf_template_dict(patient: bool = True) -> Dict[str, str]:
    return {
        'css': pdf_css(patient),
        'PDF_LOGO_ABS_URL': settings.PDF_LOGO_ABS_URL,
        'PDF_LOGO_WIDTH': settings.PDF_LOGO_WIDTH,
        'TRAFFIC_LIGHT_RED_ABS_URL': settings.TRAFFIC_LIGHT_RED_ABS_URL,
        'TRAFFIC_LIGHT_YELLOW_ABS_URL': settings.TRAFFIC_LIGHT_YELLOW_ABS_URL,
        'TRAFFIC_LIGHT_GREEN_ABS_URL': settings.TRAFFIC_LIGHT_GREEN_ABS_URL,
    }


def render_pdf_html_to_string(template: str,
                              context: Dict[str, Any] = None,
                              patient: bool = True) -> str:
    context = context or {}
    context.update(pdf_template_dict(patient))
    return render_to_string(template, context)


def email_css() -> str:
    contents = read_static_file_contents('base.css')
    contents += render_to_string('email.css')
    return contents


@django_cache_function(timeout=None)
# @lru_cache(maxsize=None)
def email_template_dict() -> Dict[str, str]:
    return {
        'css': email_css(),
    }


def render_email_html_to_string(template: str,
                                context: Dict[str, Any] = None) -> str:
    context = context or {}
    context.update(email_template_dict())
    return render_to_string(template, context)


def get_domain_from_email(email: str) -> str:
    # Very simple version...
    try:
        return email.split('@')[1]
    except:
        raise ValidationError("Bad e-mail address: no domain")


def validate_researcher_email_domain(email: str) -> None:
    if not settings.VALID_RESEARCHER_EMAIL_DOMAINS:
        # Anything goes.
        return
    domain = get_domain_from_email(email)
    for valid_domain in settings.VALID_RESEARCHER_EMAIL_DOMAINS:
        if domain.lower() == valid_domain.lower():
            return
    raise ValidationError("Invalid researcher e-mail domain")


def days_to_years(days: int, dp: int = 1) -> str:
    """
    For "consent after discharge", primarily.
    Returns the number of years to specified number of dp.
    Assumes 365 days/year, not 365.24.
    """
    try:
        years = days / 365
        if years % 1:  # needs decimals
            return "{:.{precision}f}".format(years, precision=dp)
        else:
            return str(int(years))
    except (TypeError, ValueError):
        return "?"


def latest_date(*args) -> Optional[datetime.date]:
    latest = None
    for d in args:
        if d is None:
            continue
        if latest is None:
            latest = d
        else:
            latest = max(d, latest)
    return latest



#!/usr/bin/env python
# crate_anon/nlp_manager/parse_clinical.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
import sys
from typing import Any, Dict, Generator, List, Optional, Tuple

from sqlalchemy import Column, Integer, Float, String, Text

from crate_anon.nlp_manager.nlp_definition import NlpDefinition
from crate_anon.nlp_manager.regex_parser import (
    BaseNlpParser,
    common_tense,
    compile_regex,
    FN_CONTENT,
    FN_END,
    FN_RELATION,
    FN_RELATION_TEXT,
    FN_START,
    FN_TENSE,
    FN_TENSE_TEXT,
    FN_UNITS,
    FN_VALUE_TEXT,
    FN_VARIABLE_NAME,
    FN_VARIABLE_TEXT,
    HELP_CONTENT,
    HELP_END,
    HELP_RELATION,
    HELP_RELATION_TEXT,
    HELP_START,
    HELP_TENSE,
    HELP_UNITS,
    HELP_VALUE_TEXT,
    HELP_VARIABLE_TEXT,
    MAX_RELATION_LENGTH,
    MAX_RELATION_TEXT_LENGTH,
    MAX_TENSE_LENGTH,
    MAX_UNITS_LENGTH,
    MAX_VALUE_TEXT_LENGTH,
    NumericalResultParser,
    OPTIONAL_RESULTS_IGNORABLES,
    RELATION,
    SimpleNumericalResultParser,
    TENSE_INDICATOR,
    to_float,
    to_pos_float,
    ValidatorBase,
    WORD_BOUNDARY,
)
from crate_anon.nlp_manager.regex_numbers import SIGNED_FLOAT
from crate_anon.nlp_manager.regex_units import (
    assemble_units,
    CM,
    FEET,
    INCHES,
    KG,
    kg_from_st_lb_oz,
    KG_PER_SQ_M,
    LB,
    M,
    m_from_ft_in,
    m_from_m_cm,
    MM_HG,
    STONES,
)

log = logging.getLogger(__name__)


# =============================================================================
#  Anthropometrics
# =============================================================================

# -----------------------------------------------------------------------------
# Height
# -----------------------------------------------------------------------------

class Height(NumericalResultParser):
    """Height. Handles metric and imperial."""
    METRIC_HEIGHT = r"""
        (                           # capture group 4
            (?:
                ( {SIGNED_FLOAT} )          # capture group 5
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {M} )                     # capture group 6
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {SIGNED_FLOAT} )          # capture group 7
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {CM} )                    # capture group 8
            )
            | (?:
                ( {SIGNED_FLOAT} )          # capture group 9
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {M} )                     # capture group 10
            )
            | (?:
                ( {SIGNED_FLOAT} )          # capture group 11
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {CM} )                    # capture group 12
            )
        )
    """.format(
        SIGNED_FLOAT=SIGNED_FLOAT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        M=M,
        CM=CM
    )
    IMPERIAL_HEIGHT = r"""
        (                           # capture group 13
            (?:
                ( {SIGNED_FLOAT} )      # capture group 14
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {FEET} )              # capture group 15
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {SIGNED_FLOAT} )      # capture group 16
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {INCHES} )            # capture group 17
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group 18
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {FEET} )              # capture group 19
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group 20
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {INCHES} )            # capture group 21
            )
        )
    """.format(
        SIGNED_FLOAT=SIGNED_FLOAT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        FEET=FEET,
        INCHES=INCHES
    )
    HEIGHT = r"(?: \b height \b)"
    REGEX = r"""
        ( {HEIGHT} )                       # group 1 for "height" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group 2 for tense
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group 3 for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        (?:
            {METRIC_HEIGHT}
            | {IMPERIAL_HEIGHT}
        )
    """.format(
        HEIGHT=HEIGHT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        METRIC_HEIGHT=METRIC_HEIGHT,
        IMPERIAL_HEIGHT=IMPERIAL_HEIGHT,
        SIGNED_FLOAT=SIGNED_FLOAT,
        KG_PER_SQ_M=KG_PER_SQ_M,
    )

    COMPILED_REGEX = compile_regex(REGEX)
    NAME = "Height"
    PREFERRED_UNIT_COLUMN = "value_m"

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False,
                 debug: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            regex_str_for_debugging=self.REGEX,
            commit=commit
        )
        if debug:
            print("Regex for {}: {}".format(type(self).__name__, self.REGEX))

    def parse(self, text: str,
              debug: bool = False) -> Generator[Tuple[str, Dict[str, Any]],
                                                None, None]:
        """Parser for Height. Specialized for complex unit conversion."""
        for m in self.COMPILED_REGEX.finditer(text):  # watch out: 'm'/metres
            if debug:
                print("Match {} for {}".format(m, repr(text)))
            startpos = m.start()
            endpos = m.end()
            matching_text = m.group(0)  # the whole thing
            variable_text = m.group(1)
            tense_text = m.group(2)
            relation_text = m.group(3)
            metric_expression = m.group(4)
            metric_m_and_cm_m = m.group(5)
            metric_m_and_cm_m_units = m.group(6)
            metric_m_and_cm_cm = m.group(7)
            metric_m_and_cm_cm_units = m.group(8)
            metric_m_only_m = m.group(9)
            metric_m_only_m_units = m.group(10)
            metric_cm_only_cm = m.group(11)
            metric_cm_only_cm_units = m.group(12)
            imperial_expression = m.group(13)
            imperial_ft_and_in_ft = m.group(14)
            imperial_ft_and_in_ft_units = m.group(15)
            imperial_ft_and_in_in = m.group(16)
            imperial_ft_and_in_in_units = m.group(17)
            imperial_ft_only_ft = m.group(18)
            imperial_ft_only_ft_units = m.group(19)
            imperial_in_only_in = m.group(20)
            imperial_in_only_in_units = m.group(21)

            expression = None
            value_m = None
            units = None
            if metric_expression:
                expression = metric_expression
                if metric_m_and_cm_m and metric_m_and_cm_cm:
                    metres = to_pos_float(metric_m_and_cm_m)
                    # ... beware: 'm' above
                    cm = to_pos_float(metric_m_and_cm_cm)
                    value_m = m_from_m_cm(metres=metres, centimetres=cm)
                    units = assemble_units([metric_m_and_cm_m_units,
                                            metric_m_and_cm_cm_units])
                elif metric_m_only_m:
                    value_m = to_pos_float(metric_m_only_m)
                    units = metric_m_only_m_units
                elif metric_cm_only_cm:
                    cm = to_pos_float(metric_cm_only_cm)
                    value_m = m_from_m_cm(centimetres=cm)
                    units = metric_cm_only_cm_units
            elif imperial_expression:
                expression = imperial_expression
                if imperial_ft_and_in_ft and imperial_ft_and_in_in:
                    ft = to_pos_float(imperial_ft_and_in_ft)
                    inches = to_pos_float(imperial_ft_and_in_in)
                    value_m = m_from_ft_in(feet=ft, inches=inches)
                    units = assemble_units([imperial_ft_and_in_ft_units,
                                            imperial_ft_and_in_in_units])
                elif imperial_ft_only_ft:
                    ft = to_pos_float(imperial_ft_only_ft)
                    value_m = m_from_ft_in(feet=ft)
                    units = imperial_ft_only_ft_units
                elif imperial_in_only_in:
                    inches = to_pos_float(imperial_in_only_in)
                    value_m = m_from_ft_in(inches=inches)
                    units = imperial_in_only_in_units

            tense, relation = common_tense(tense_text, relation_text)

            result = {
                FN_VARIABLE_NAME: self.variable,
                FN_CONTENT: matching_text,
                FN_START: startpos,
                FN_END: endpos,

                FN_VARIABLE_TEXT: variable_text,
                FN_RELATION_TEXT: relation_text,
                FN_RELATION: relation,
                FN_VALUE_TEXT: expression,
                FN_UNITS: units,
                self.target_unit: value_m,
                FN_TENSE_TEXT: tense_text,
                FN_TENSE: tense,
            }
            # log.critical(result)
            yield self.tablename, result

    def test(self, verbose: bool = False) -> None:
        self.test_numerical_parser([
            ("Height", []),  # should fail; no values
            ("her height was 1.6m", [1.6]),
            ("Height = 1.23 m", [1.23]),
            ("her height is 1.5m", [1.5]),
            ('''Height 5'8" ''', [m_from_ft_in(feet=5, inches=8)]),
            ("Height 5 ft 8 in", [m_from_ft_in(feet=5, inches=8)]),
            ("Height 5 feet 8 inches", [m_from_ft_in(feet=5, inches=8)]),
        ], verbose=verbose)
        self.detailed_test("Height 5 ft 11 in", [{
            self.target_unit: m_from_ft_in(feet=5, inches=11),
            FN_UNITS: "ft in",
        }], verbose=verbose)
        # *** Height NLP: deal with "tall" and plain "is", e.g.
        # she is 6'2"; she is 1.5m tall


class HeightValidator(ValidatorBase):
    """Validator for Height (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Height.HEIGHT],
                         validated_variable=Height.NAME,
                         commit=commit)


# -----------------------------------------------------------------------------
# Weight (mass)
# -----------------------------------------------------------------------------

class Weight(NumericalResultParser):
    """Weight. Handles metric and imperial."""
    METRIC_WEIGHT = r"""
        (                           # capture group 4
            ( {SIGNED_FLOAT} )          # capture group 5
            {OPTIONAL_RESULTS_IGNORABLES}
            ( {KG} )                    # capture group 6
        )
    """.format(
        SIGNED_FLOAT=SIGNED_FLOAT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        KG=KG
    )
    IMPERIAL_WEIGHT = r"""
        (                           # capture group 7
            (?:
                ( {SIGNED_FLOAT} )      # capture group 8
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {STONES} )            # capture group 9
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {SIGNED_FLOAT} )      # capture group 10
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {LB} )                # capture group 11
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group 12
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {STONES} )            # capture group 13
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group 14
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {LB} )                # capture group 15
            )
        )
    """.format(
        SIGNED_FLOAT=SIGNED_FLOAT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        STONES=STONES,
        LB=LB
    )
    WEIGHT = r"(?: \b weigh[ts] \b )"  # weight, weighs
    REGEX = r"""
        ( {WEIGHT} )                       # group 1 for "weight" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group 2 for tense
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group 3 for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        (?:
            {METRIC_WEIGHT}
            | {IMPERIAL_WEIGHT}
        )
    """.format(
        WEIGHT=WEIGHT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        METRIC_WEIGHT=METRIC_WEIGHT,
        IMPERIAL_WEIGHT=IMPERIAL_WEIGHT,
        SIGNED_FLOAT=SIGNED_FLOAT,
        KG_PER_SQ_M=KG_PER_SQ_M,
    )

    COMPILED_REGEX = compile_regex(REGEX)
    NAME = "Weight"
    PREFERRED_UNIT_COLUMN = "value_kg"

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False,
                 debug: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            regex_str_for_debugging=self.REGEX,
            commit=commit
        )
        if debug:
            print("Regex for {}: {}".format(type(self).__name__, self.REGEX))

    def parse(self, text: str,
              debug: bool = False) -> Generator[Tuple[str, Dict[str, Any]],
                                                None, None]:
        """Parser for Weight. Specialized for complex unit conversion."""
        for m in self.COMPILED_REGEX.finditer(text):
            if debug:
                print("Match {} for {}".format(m, repr(text)))
            startpos = m.start()
            endpos = m.end()
            matching_text = m.group(0)  # the whole thing
            variable_text = m.group(1)
            tense_text = m.group(2)
            relation_text = m.group(3)
            metric_expression = m.group(4)
            metric_value = m.group(5)
            metric_units = m.group(6)
            imperial_expression = m.group(7)
            imperial_st_and_lb_st = m.group(8)
            imperial_st_and_lb_st_units = m.group(9)
            imperial_st_and_lb_lb = m.group(10)
            imperial_st_and_lb_lb_units = m.group(11)
            imperial_st_only_st = m.group(12)
            imperial_st_only_st_units = m.group(13)
            imperial_lb_only_lb = m.group(14)
            imperial_lb_only_lb_units = m.group(15)

            expression = None
            value_kg = None
            units = None
            if metric_expression:
                expression = metric_expression
                value_kg = to_float(metric_value)
                units = metric_units
            elif imperial_expression:
                expression = imperial_expression
                if imperial_st_and_lb_st and imperial_st_and_lb_lb:
                    st = to_float(imperial_st_and_lb_st)
                    lb = to_float(imperial_st_and_lb_lb)
                    value_kg = kg_from_st_lb_oz(stones=st, pounds=lb)
                    units = assemble_units([imperial_st_and_lb_st_units,
                                            imperial_st_and_lb_lb_units])
                elif imperial_st_only_st:
                    st = to_float(imperial_st_only_st)
                    value_kg = kg_from_st_lb_oz(stones=st)
                    units = imperial_st_only_st_units
                elif imperial_lb_only_lb:
                    lb = to_float(imperial_lb_only_lb)
                    value_kg = kg_from_st_lb_oz(pounds=lb)
                    units = imperial_lb_only_lb_units

            # All left as signed float, as you definitely see things like
            # "weight -0.3 kg" for weight changes.

            tense, relation = common_tense(tense_text, relation_text)

            result = {
                FN_VARIABLE_NAME: self.variable,
                FN_CONTENT: matching_text,
                FN_START: startpos,
                FN_END: endpos,

                FN_VARIABLE_TEXT: variable_text,
                FN_RELATION_TEXT: relation_text,
                FN_RELATION: relation,
                FN_VALUE_TEXT: expression,
                FN_UNITS: units,
                self.target_unit: value_kg,
                FN_TENSE_TEXT: tense_text,
                FN_TENSE: tense,
            }
            # log.critical(result)
            yield self.tablename, result

    def test(self, verbose: bool = False) -> None:
        self.test_numerical_parser([
            ("Weight", []),  # should fail; no values
            ("her weight was 60.2kg", [60.2]),
            ("Weight = 52.3kg", [52.3]),
            ("Weight: 80.8kgs", [80.8]),
            ("she weighs 61kg", [61]),
            ("she weighs 61 kg", [61]),
            ("she weighs 61 kgs", [61]),
            ("she weighs 61 kilo", [61]),
            ("she weighs 61 kilos", [61]),
            ("she weighs 8 stones ", [kg_from_st_lb_oz(stones=8)]),
            ("she weighs 200 lb", [kg_from_st_lb_oz(pounds=200)]),
            ("she weighs 200 pounds", [kg_from_st_lb_oz(pounds=200)]),
            ("she weighs 6 st 12 lb", [kg_from_st_lb_oz(stones=6, pounds=12)]),
            ("change in weight -0.4kg", [-0.4]),
            ("change in weight - 0.4kg", [0.4]),  # ASCII hyphen (hyphen-minus)
            ("change in weight ‐ 0.4kg", [0.4]),  # Unicode hyphen
            # ("failme", [999]),
            ("change in weight −0.4kg", [-0.4]),  # Unicode minus
            ("change in weight –0.4kg", [-0.4]),  # en dash
            ("change in weight —0.4kg", [0.4]),  # em dash
        ], verbose=verbose)
        self.detailed_test("Weight: 80.8kgs", [{
            self.target_unit: 80.8,
            FN_UNITS: "kgs",
        }], verbose=verbose)


class WeightValidator(ValidatorBase):
    """Validator for Weight (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Weight.WEIGHT],
                         validated_variable=Weight.NAME,
                         commit=commit)


# -----------------------------------------------------------------------------
# Body mass index (BMI)
# -----------------------------------------------------------------------------

class Bmi(SimpleNumericalResultParser):
    """Body mass index (in kg / m^2)."""
    BMI = r"""
        (?: {WORD_BOUNDARY}
            (?: BMI | body \s+ mass \s+ index )
        {WORD_BOUNDARY} )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    REGEX = r"""
        ( {BMI} )                          # group for "BMI" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {SIGNED_FLOAT} )                 # group for value
        {OPTIONAL_RESULTS_IGNORABLES}
        (                                  # group for units
            {KG_PER_SQ_M}
        )?
    """.format(
        BMI=BMI,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        KG_PER_SQ_M=KG_PER_SQ_M,
    )
    NAME = "BMI"
    PREFERRED_UNIT_COLUMN = "value_kg_per_sq_m"
    UNIT_MAPPING = {
        KG_PER_SQ_M: 1,       # preferred unit
    }
    # deal with "a BMI of 30"?

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            regex_str=self.REGEX,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            units_to_factor=self.UNIT_MAPPING,
            commit=commit,
            take_absolute=True
        )

    def test(self, verbose: bool = False) -> None:
        self.test_numerical_parser([
            ("BMI", []),  # should fail; no values
            ("body mass index was 30", [30]),
            ("his BMI (30) is too high", [30]),
            ("BMI 25 kg/sq m", [25]),
            ("BMI was 18.4 kg/m^-2", [18.4]),
            ("ACE 79", []),
            ("BMI-23", [23]),
        ], verbose=verbose)


class BmiValidator(ValidatorBase):
    """Validator for Bmi (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Bmi.BMI],
                         validated_variable=Bmi.NAME,
                         commit=commit)


# =============================================================================
#  Bedside investigations: BP
# =============================================================================

class Bp(BaseNlpParser):
    """Blood pressure, in mmHg. (Since we produce two variables, SBP and DBP,
    and we use something a little more complex than
    NumeratorOutOfDenominatorParser, we subclass BaseNlpParser directly.)"""
    BP = r"(?: \b blood \s+ pressure \b | \b B\.?P\.? \b )"
    SYSTOLIC_BP = r"(?: \b systolic \s+ {BP} | \b S\.?B\.?P\.? \b )".format(
        BP=BP)
    DIASTOLIC_BP = r"(?: \b diastolic \s+ {BP} | \b D\.?B\.?P\.? \b )".format(
        BP=BP)

    TWO_NUMBER_BP = r"""
        ( {SIGNED_FLOAT} )
        \s* (?: \b over \b | \/ ) \s*
        ( {SIGNED_FLOAT} )
    """.format(SIGNED_FLOAT=SIGNED_FLOAT)
    ONE_NUMBER_BP = SIGNED_FLOAT

    COMPILED_BP = compile_regex(BP)
    COMPILED_SBP = compile_regex(SYSTOLIC_BP)
    COMPILED_DBP = compile_regex(DIASTOLIC_BP)
    COMPILED_ONE_NUMBER_BP = compile_regex(ONE_NUMBER_BP)
    COMPILED_TWO_NUMBER_BP = compile_regex(TWO_NUMBER_BP)
    REGEX = r"""
        (                               # group for "BP" or equivalent
            {SYSTOLIC_BP}               # ... from more to less specific
            | {DIASTOLIC_BP}
            | {BP}
        )
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?         # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        (
            {SIGNED_FLOAT}                      # systolic
            (?:
                \s* (?: \b over \b | \/ ) \s*   # /
                {SIGNED_FLOAT}                  # diastolic
            )?
        )
        {OPTIONAL_RESULTS_IGNORABLES}
        (                              # group for units
            {MM_HG}
        )?
    """.format(
        BP=BP,
        SYSTOLIC_BP=SYSTOLIC_BP,
        DIASTOLIC_BP=DIASTOLIC_BP,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        MM_HG=MM_HG,
    )
    COMPILED_REGEX = compile_regex(REGEX)

    FN_SYSTOLIC_BP_MMHG = 'systolic_bp_mmhg'
    FN_DIASTOLIC_BP_MMHG = 'diastolic_bp_mmhg'

    NAME = "BP"
    UNIT_MAPPING = {
        MM_HG: 1,       # preferred unit
    }

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            commit=commit
        )
        if nlpdef is None:  # only None for debugging!
            self.tablename = ''
        else:
            self.tablename = nlpdef.opt_str(
                cfgsection, 'desttable', required=True)

    @classmethod
    def print_info(cls, file=sys.stdout):
        print("Blood pressure finder. Regular expression: \n{}".format(
            cls.REGEX), file=file)

    def dest_tables_columns(self) -> Dict[str, List[Column]]:
        return {self.tablename: [
            Column(FN_CONTENT, Text, doc=HELP_CONTENT),
            Column(FN_START, Integer, doc=HELP_START),
            Column(FN_END, Integer, doc=HELP_END),
            Column(FN_VARIABLE_TEXT, Text, doc=HELP_VARIABLE_TEXT),
            Column(FN_RELATION_TEXT, String(MAX_RELATION_TEXT_LENGTH),
                   doc=HELP_RELATION_TEXT),
            Column(FN_RELATION, String(MAX_RELATION_LENGTH),
                   doc=HELP_RELATION),
            Column(FN_VALUE_TEXT, String(MAX_VALUE_TEXT_LENGTH),
                   doc=HELP_VALUE_TEXT),
            Column(FN_UNITS, String(MAX_UNITS_LENGTH), doc=HELP_UNITS),
            Column(self.FN_SYSTOLIC_BP_MMHG, Float,
                   doc="Systolic blood pressure in mmHg"),
            Column(self.FN_DIASTOLIC_BP_MMHG, Float,
                   doc="Diastolic blood pressure in mmHg"),
            Column(FN_TENSE, String(MAX_TENSE_LENGTH), doc=HELP_TENSE),
        ]}

    def parse(self, text: str,
              debug: bool = False) -> Generator[Tuple[str, Dict[str, Any]],
                                                None, None]:
        """Parser for BP. Specialized because we're fetching two numbers."""
        for m in self.COMPILED_REGEX.finditer(text):
            if debug:
                print("Match {} for {}".format(m, repr(text)))
            startpos = m.start()
            endpos = m.end()
            matching_text = m.group(0)  # the whole thing
            variable_text = m.group(1)
            tense_indicator = m.group(2)
            relation_text = m.group(3)
            value_text = m.group(4)
            units = m.group(5)

            sbp = None
            dbp = None
            if self.COMPILED_SBP.match(variable_text):
                if self.COMPILED_ONE_NUMBER_BP.match(value_text):
                    sbp = to_pos_float(value_text)
            elif self.COMPILED_DBP.match(variable_text):
                if self.COMPILED_ONE_NUMBER_BP.match(value_text):
                    dbp = to_pos_float(value_text)
            elif self.COMPILED_BP.match(variable_text):
                bpmatch = self.COMPILED_TWO_NUMBER_BP.match(value_text)
                if bpmatch:
                    sbp = to_pos_float(bpmatch.group(1))
                    dbp = to_pos_float(bpmatch.group(2))
            if sbp is None and dbp is None:
                # This is OK; e.g. "BP 110", which we will ignore.
                # log.warning(
                #     "Failed interpretation: matching_text={matching_text}, "
                #     "variable_text={variable_text}, "
                #     "tense_indicator={tense_indicator}, "
                #     "relation={relation}, "
                #     "value_text={value_text}, "
                #     "units={units}".format(
                #         matching_text=repr(matching_text),
                #         variable_text=repr(variable_text),
                #         tense_indicator=repr(tense_indicator),
                #         relation=repr(relation),
                #         value_text=repr(value_text),
                #         units=repr(units),
                #     )
                # )
                continue

            tense, relation = common_tense(tense_indicator, relation_text)

            yield self.tablename, {
                FN_CONTENT: matching_text,
                FN_START: startpos,
                FN_END: endpos,
                FN_VARIABLE_TEXT: variable_text,
                FN_RELATION_TEXT: relation_text,
                FN_RELATION: relation,
                FN_VALUE_TEXT: value_text,
                FN_UNITS: units,
                self.FN_SYSTOLIC_BP_MMHG: sbp,
                self.FN_DIASTOLIC_BP_MMHG: dbp,
                FN_TENSE: tense,
            }

    def test_bp_parser(
            self,
            test_expected_list: List[
                Tuple[str, List[Tuple[float, float]]]
            ],
            verbose: bool = False) -> None:
        print("Testing parser: {}".format(type(self).__name__))
        if verbose:
            print("... regex:\n{}".format(self.REGEX))
        for test_string, expected_values in test_expected_list:
            actual_values = list(
                (x[self.FN_SYSTOLIC_BP_MMHG], x[self.FN_DIASTOLIC_BP_MMHG])
                for t, x in self.parse(test_string)
            )
            assert actual_values == expected_values, (
                "Parser {name}: Expected {expected}, got {actual}, when "
                "parsing {test_string}; full result={full}".format(
                    name=type(self).__name__,
                    expected=expected_values,
                    actual=actual_values,
                    test_string=repr(test_string),
                    full=repr(list(self.parse(test_string))),
                )
            )
        print("... OK")

    def test(self, verbose: bool = False) -> None:
        self.test_bp_parser([
            ("BP", []),  # should fail; no values
            ("his blood pressure was 120/80", [(120, 80)]),
            ("BP 120/80 mmhg", [(120, 80)]),
            ("systolic BP 120", [(120, None)]),
            ("diastolic BP 80", [(None, 80)]),
            ("BP-130/70", [(130, 70)]),
            ("BP 110 /80", [(110, 80)]),
            ("BP 110 /80 -", [(110, 80)]),  # real example
            ("BP 120 / 70 -", [(120, 70)]),  # real example
            ("BP :115 / 70 -", [(115, 70)]),  # real example
            ("B.P 110", []),  # real example
        ], verbose=verbose)
        # 1. Unsure if best to take abs value.
        #    One reason not to might be if people express changes, e.g.
        #    "BP change -40/-10", but I very much doubt it.
        #    Went with abs value using to_pos_float().
        # 2. "BP 110" - too unreliable; not definitely a blood pressure.


class BpValidator(ValidatorBase):
    """Validator for Bp (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Bp.REGEX],
                         validated_variable=Bp.NAME,
                         commit=commit)


# =============================================================================
#  Command-line entry point
# =============================================================================

def test_all(verbose: bool = False) -> None:
    height = Height(None, None)
    height.test(verbose=verbose)
    weight = Weight(None, None)
    weight.test(verbose=verbose)
    bmi = Bmi(None, None)
    bmi.test(verbose=verbose)
    bp = Bp(None, None)
    bp.test(verbose=verbose)


if __name__ == '__main__':
    test_all(verbose=True)



##!/usr/bin/env python
# crate_anon/nlp_manager/input_field_config.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
# import sys
from typing import Any, Dict, Generator, List, Optional, Tuple

from cardinal_pythonlib.rnc_db import (
    ensure_valid_field_name,
    ensure_valid_table_name,
)
from sqlalchemy import BigInteger, Column, Index, String, Table
from sqlalchemy.sql import and_, column, exists, or_, select, table

from crate_anon.nlp_manager.constants import (
    FN_NLPDEF,
    FN_PK,
    FN_SRCDB,
    FN_SRCTABLE,
    FN_SRCPKFIELD,
    FN_SRCPKVAL,
    FN_SRCPKSTR,
    FN_SRCFIELD,
    MAX_STRING_PK_LENGTH,
)
from crate_anon.common.timing import MultiTimerContext, timer
from crate_anon.common.hash import hash64
from crate_anon.common.parallel import is_my_job_by_hash
from crate_anon.common.sqla import (
    count_star,
    is_sqlatype_integer,
    get_column_type,
    table_exists,
)
from crate_anon.nlp_manager.constants import SqlTypeDbIdentifier
from crate_anon.nlp_manager.models import NlpRecord

# if sys.version_info.major >= 3 and sys.version_info.minor >= 5:
#     from crate_anon.nlp_manager import nlp_definition  # see PEP0484
from crate_anon.nlp_manager.nlp_definition import NlpDefinition

log = logging.getLogger(__name__)

TIMING_GEN_TEXT_SQL_SELECT = "gen_text_sql_select"
TIMING_PROCESS_GEN_TEXT = "process_generated_text"
TIMING_PROGRESS_DB_SELECT = "progress_db_select"
TIMING_PROGRESS_DB_DELETE = "progress_db_delete"


# =============================================================================
# Input field definition
# =============================================================================

class InputFieldConfig(object):
    """
    Class defining configuration for an input field (containing text).
    """

    def __init__(self, nlpdef: NlpDefinition, section: str) -> None:
        """
        Read config from a configparser section.
        """
        def opt_str(option: str) -> str:
            return nlpdef.opt_str(section, option, required=True)

        def opt_strlist(option: str,
                        required: bool = False,
                        lower: bool = True) -> List[str]:
            return nlpdef.opt_strlist(section, option, as_words=False,
                                      lower=lower, required=required)

        def opt_int(option: str, default: Optional[int]) -> Optional[int]:
            return nlpdef.opt_int(section, option, default=default)

        # def opt_bool(option: str, default: bool) -> bool:
        #     return nlpdef.opt_bool(section, option, default=default)

        self._nlpdef = nlpdef

        self._srcdb = opt_str('srcdb')
        self._srctable = opt_str('srctable')
        self._srcpkfield = opt_str('srcpkfield')
        self._srcfield = opt_str('srcfield')
        # Make these case-sensitive to avoid our failure in renaming SQLA
        # Column objects to be lower-case:
        self._copyfields = opt_strlist('copyfields', lower=False)  # fieldnames
        self._indexed_copyfields = opt_strlist('indexed_copyfields',
                                               lower=False)
        self._debug_row_limit = opt_int('debug_row_limit', default=0)
        # self._fetch_sorted = opt_bool('fetch_sorted', default=True)

        ensure_valid_table_name(self._srctable)
        ensure_valid_field_name(self._srcpkfield)
        ensure_valid_field_name(self._srcfield)

        if len(set(self._indexed_copyfields)) != len(self._indexed_copyfields):
            raise ValueError("Redundant indexed_copyfields: {}".format(
                self._indexed_copyfields))

        if len(set(self._copyfields)) != len(self._copyfields):
            raise ValueError("Redundant copyfields: {}".format(
                self._copyfields))

        indexed_not_copied = set(self._indexed_copyfields) - set(
            self._copyfields)
        if indexed_not_copied:
            raise ValueError("Fields in index_copyfields but not in "
                             "copyfields: {}".format(indexed_not_copied))

        # allfields = [self._srcpkfield, self._srcfield] + self._copyfields
        # if len(allfields) != len(set(allfields)):
        #     raise ValueError(
        #         "Field overlap in InputFieldConfig: {}".format(section))
        # RE-THOUGHT: OK to copy source text fields etc. if desired.
        # It's fine in SQL to say SELECT a, a FROM mytable;

        self._db = nlpdef.get_database(self._srcdb)

    def get_srcdb(self) -> str:
        return self._srcdb

    def get_srctable(self) -> str:
        return self._srctable

    def get_srcpkfield(self) -> str:
        return self._srcpkfield

    def get_srcfield(self) -> str:
        return self._srcfield

    def get_source_session(self):
        return self._db.session

    def _get_source_metadata(self):
        return self._db.metadata

    def _get_source_engine(self):
        return self._db.engine

    def _get_progress_session(self):
        return self._nlpdef.get_progdb_session()

    @staticmethod
    def get_core_columns_for_dest() -> List[Column]:
        """Core columns in destination tables, primarily referring to the
        source."""
        return [
            Column(FN_PK, BigInteger, primary_key=True,
                   autoincrement=True,
                   doc="Arbitrary primary key (PK) of output record"),
            Column(FN_NLPDEF, SqlTypeDbIdentifier,
                   doc="Name of the NLP definition producing this row"),
            Column(FN_SRCDB, SqlTypeDbIdentifier,
                   doc="Source database name (from CRATE NLP config)"),
            Column(FN_SRCTABLE, SqlTypeDbIdentifier,
                   doc="Source table name"),
            Column(FN_SRCPKFIELD, SqlTypeDbIdentifier,
                   doc="PK field (column) name in source table"),
            Column(FN_SRCPKVAL, BigInteger,
                   doc="PK of source record (or integer hash of PK if the PK "
                       "is a string)"),
            Column(FN_SRCPKSTR, String(MAX_STRING_PK_LENGTH),
                   doc="NULL if the table has an integer PK, but the PK if "
                       "the PK was a string, to deal with hash collisions. "
                       "Max length: {}".format(MAX_STRING_PK_LENGTH)),
            Column(FN_SRCFIELD, SqlTypeDbIdentifier,
                   doc="Field (column) name of source text"),
        ]

    @staticmethod
    def get_core_indexes_for_dest() -> List[Index]:
        """Indexes for columns, primarily referring to the source."""
        # http://stackoverflow.com/questions/179085/multiple-indexes-vs-multi-column-indexes  # noqa
        return [
            Index('_idx_srcref',
                  # Remember, order matters; more to less specific
                  # See also BaseNlpParser.delete_dest_record
                  FN_SRCPKVAL,
                  FN_NLPDEF,
                  FN_SRCFIELD,
                  FN_SRCTABLE,
                  FN_SRCDB,
                  FN_SRCPKSTR),
            Index('_idx_deletion',
                  # We sometimes delete just using the following; see
                  # BaseNlpParser.delete_where_srcpk_not
                  FN_NLPDEF,
                  FN_SRCFIELD,
                  FN_SRCTABLE,
                  FN_SRCDB),
        ]

    def _require_table_exists(self) -> None:
        if not table_exists(self._get_source_engine(), self._srctable):
            msg = "Missing source table: {}.{}".format(self._srcdb,
                                                       self._srctable)
            log.critical(msg)
            raise ValueError(msg)

    def get_copy_columns(self) -> List[Column]:
        # We read the column type from the source database.
        self._require_table_exists()
        meta = self._get_source_metadata()
        t = Table(self._srctable, meta, autoload=True)
        copy_columns = []  # type: List[Column]
        processed_copy_column_names = []  # type: List[str]
        for c in t.columns:
            # if c.name.lower() in self._copyfields:
            if c.name in self._copyfields:
                copied = c.copy()
                # Force lower case:
                # copied.name = copied.name.lower()
                # copied.name = quoted_name(copied.name.lower(), None)
                # ... this is not working properly. Keep getting an
                # "Unconsumed column names" error with e.g. a source field of
                # "Text".
                # Try making copyfields case-sensitive instead.
                copy_columns.append(copied)
                processed_copy_column_names.append(c.name)
        # Check all requested fields are present:
        missing = set(self._copyfields) - set(processed_copy_column_names)
        if missing:
            raise ValueError(
                "The following fields were requested to be copied but are "
                "absent from the source (NB case-sensitive): {}".format(
                    missing))
        # log.critical(copy_columns)
        return copy_columns

    def get_copy_indexes(self) -> List[Index]:
        self._require_table_exists()
        meta = self._get_source_metadata()
        t = Table(self._srctable, meta, autoload=True)
        copy_indexes = []  # type: List[Index]
        processed_copy_index_col_names = []  # type: List[str]
        for c in t.columns:
            # if c.name.lower() in self._indexed_copyfields:
            if c.name in self._indexed_copyfields:
                copied = c.copy()
                # See above re case.
                idx_name = "idx_{}".format(c.name)
                copy_indexes.append(Index(idx_name, copied))
                processed_copy_index_col_names.append(c.name)
        missing = set(self._indexed_copyfields) - set(
            processed_copy_index_col_names)
        if missing:
            raise ValueError(
                "The following fields were requested to be copied/indexed but "
                "are absent from the source (NB case-sensitive): {}".format(
                    missing))
        return copy_indexes

    def is_pk_integer(self):
        pkcoltype = get_column_type(self._get_source_engine(), self._srctable,
                                    self._srcpkfield)
        if not pkcoltype:
            raise ValueError("Unable to get column type for column "
                             "{}.{}".format(self._srctable, self._srcpkfield))
        pk_is_integer = is_sqlatype_integer(pkcoltype)
        # log.debug("pk_is_integer: {} -> {}".format(repr(pkcoltype),
        #                                            pk_is_integer))
        return pk_is_integer

    def gen_text(self, tasknum: int = 0,
                 ntasks: int = 1) -> Generator[Tuple[str, Dict[str, Any]],
                                               None, None]:
        """
        Generate text strings from the input database.
        Yields tuple of (text, dict), where the dict is a column-to-value
        mapping for all other fields (source reference fields, copy fields).
        """
        if 1 < ntasks <= tasknum:
            raise Exception("Invalid tasknum {}; must be <{}".format(
                tasknum, ntasks))
        base_dict = {
            FN_SRCDB: self._srcdb,
            FN_SRCTABLE: self._srctable,
            FN_SRCPKFIELD: self._srcpkfield,
            FN_SRCFIELD: self._srcfield,
        }
        session = self.get_source_session()
        pkcol = column(self._srcpkfield)
        # ... don't use is_sqlatype_integer with this; it's a column clause,
        # not a full column definition.
        pk_is_integer = self.is_pk_integer()

        selectcols = [pkcol, column(self._srcfield)]
        for extracol in self._copyfields:
            selectcols.append(column(extracol))
        query = select(selectcols).select_from(table(self._srctable))
        # not ordered...
        # if self._fetch_sorted:
        #     query = query.order_by(pkcol)
        distribute_by_hash = False
        if ntasks > 1:
            if pk_is_integer:
                # Integer PK, so we can be efficient and bake the parallel
                # processing work division into the SQL:
                query = query.where(pkcol % ntasks == tasknum)
            else:
                distribute_by_hash = True
        nrows_returned = 0
        hashed_pk = None  # remove warning about reference before assignment
        with MultiTimerContext(timer, TIMING_GEN_TEXT_SQL_SELECT):
            result = session.execute(query)
            for row in result:  # ... a generator itself
                with MultiTimerContext(timer, TIMING_PROCESS_GEN_TEXT):
                    pkval = row[0]
                    if (distribute_by_hash and
                            not is_my_job_by_hash(pkval, tasknum, ntasks)):
                        continue

                    if 0 < self._debug_row_limit <= nrows_returned:
                        log.warning(
                            "Table {}.{}: not fetching more than {} rows (in "
                            "total for this process) due to debugging "
                            "limits".format(self._srcdb, self._srctable,
                                            self._debug_row_limit))
                        result.close()  # http://docs.sqlalchemy.org/en/latest/core/connections.html  # noqa
                        return

                    text = row[1]
                    if not text:
                        continue
                    other_values = dict(zip(self._copyfields, row[2:]))
                    if pk_is_integer:
                        other_values[FN_SRCPKVAL] = pkval
                        other_values[FN_SRCPKSTR] = None
                    else:  # hashed_pk will have been set above
                        other_values[FN_SRCPKVAL] = hashed_pk
                        other_values[FN_SRCPKSTR] = pkval
                    other_values.update(base_dict)
                    yield text, other_values
                    nrows_returned += 1

    def get_count(self) -> int:
        """
        Counts records in the input table for the given InputFieldConfig.
        Used for progress monitoring.
        """
        return count_star(session=self.get_source_session(),
                          tablename=self._srctable)

    def get_progress_record(self,
                            srcpkval: int,
                            srcpkstr: str = None) -> Optional[NlpRecord]:
        """
        Fetch a progress record (NlpRecord) for the given source record, if one
        exists.
        """
        session = self._get_progress_session()
        query = (
            session.query(NlpRecord).
            filter(NlpRecord.srcdb == self._srcdb).
            filter(NlpRecord.srctable == self._srctable).
            filter(NlpRecord.srcpkval == srcpkval).
            filter(NlpRecord.srcfield == self._srcfield).
            filter(NlpRecord.nlpdef == self._nlpdef.get_name())
            # Order not important (though the order of the index certainly
            # is; see NlpRecord.__table_args__).
            # http://stackoverflow.com/questions/11436469/does-order-of-where-clauses-matter-in-sql  # noqa
        )
        if srcpkstr is not None:
            query = query.filter(NlpRecord.srcpkstr == srcpkstr)
        # log.critical(query)
        with MultiTimerContext(timer, TIMING_PROGRESS_DB_SELECT):
            # This was surprisingly slow under SQL Server testing.
            return query.one_or_none()

    def gen_src_pks(self) -> Generator[Tuple[int, Optional[str]], None, None]:
        """
        Generate integer PKs from the source table specified for the
        InputFieldConfig.
        Timing is subsumed under TIMING_DELETE_WHERE_NO_SOURCE.
        """
        session = self.get_source_session()
        query = (
            select([column(self._srcpkfield)]).
            select_from(table(self._srctable))
        )
        result = session.execute(query)
        if self.is_pk_integer():
            for row in result:
                yield row[0], None
        else:
            for row in result:
                pkval = row[0]
                yield hash64(pkval), pkval

    def delete_progress_records_where_srcpk_not(
            self,
            temptable: Optional[Table]) -> None:
        progsession = self._get_progress_session()
        log.debug("delete_progress_records_where_srcpk_not... {}.{} -> "
                  "progressdb".format(self._srcdb, self._srctable))
        prog_deletion_query = (
            progsession.query(NlpRecord).
            filter(NlpRecord.srcdb == self._srcdb).
            filter(NlpRecord.srctable == self._srctable).
            # unnecessary # filter(NlpRecord.srcpkfield == self._srcpkfield).
            filter(NlpRecord.nlpdef == self._nlpdef.get_name())
        )
        if temptable is not None:
            log.debug("... deleting selectively")
            temptable_pkvalcol = temptable.columns[FN_SRCPKVAL]
            temptable_pkstrcol = temptable.columns[FN_SRCPKSTR]
            prog_deletion_query = prog_deletion_query.filter(
                ~exists().where(
                    and_(
                        NlpRecord.srcpkval == temptable_pkvalcol,
                        or_(
                            NlpRecord.srcpkstr == temptable_pkstrcol,
                            and_(
                                NlpRecord.srcpkstr.is_(None),
                                temptable_pkstrcol.is_(None)
                            )
                        )
                    )
                )
            )
        else:
            log.debug("... deleting all")
        with MultiTimerContext(timer, TIMING_PROGRESS_DB_DELETE):
            prog_deletion_query.delete(synchronize_session=False)
            # http://docs.sqlalchemy.org/en/latest/orm/query.html#sqlalchemy.orm.query.Query.delete  # noqa
        self._nlpdef.commit(progsession)

    def delete_all_progress_records(self) -> None:
        progsession = self._get_progress_session()
        prog_deletion_query = (
            progsession.query(NlpRecord).
            filter(NlpRecord.nlpdef == self._nlpdef.get_name())
        )
        log.debug("delete_all_progress_records for NLP definition: {}".format(
            self._nlpdef.get_name()))
        with MultiTimerContext(timer, TIMING_PROGRESS_DB_DELETE):
            prog_deletion_query.delete(synchronize_session=False)
        self._nlpdef.commit(progsession)



#!/usr/bin/env python
# crate_anon/preprocess/rio_pk.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

__SUPERSEDED = """

RIO_6_2_ATYPICAL_PKS = {  # SUPERSEDED by better PK detection

    # These are table: pk_field mappings for PATIENT tables, i.e. those
    # containing the ClientID field, where that PK is not the default of
    # SequenceID.

    # -------------------------------------------------------------------------
    # RiO Core
    # -------------------------------------------------------------------------

    # Ams*: Appointment Management System
    'AmsAppointmentContactActivity': 'ActivitySequenceID',
    'AmsAppointmentOtherHCP': None,  # non-patient; non-unique SequenceID
    # ... SequenceID is non-unique and the docs also list it as an FK;
    #     ActivitySequenceID this is unique and a PK
    'AmsReferralDatesArchive': 'AMSSequenceID',
    # ... UNVERIFIED as no rows in our data; listed as a PK and an FK
    'AmsReferralListUrgency': None,
    'AmsReferralListWaitingStatus': None,
    'AmsStream': None,  # non-patient; non-unique SequenceID

    'CarePlanIndex': 'CarePlanID',
    'CarePlanProblemOrder': None,

    'ClientAddressMerged': None,  # disused table
    'ClientCareSpell': None,  # CareSpellNum is usually 1 for a given ClientID
    'ClientDocumentAdditionalClient': None,
    'ClientFamily': None,
    'ClientFamilyLink': None,
    'ClientGPMerged': None,
    'ClientHealthCareProvider': None,
    'ClientMerge': None,
    'ClientMerged': None,
    'ClientName': 'ClientNameID',
    'ClientOtherDetail': None,  # not in docs, but looks like Core
    'ClientPhoto': None,
    'ClientPhotoMerged': None,
    'ClientProperty': None,
    'ClientPropertyMerged': None,
    'ClientTelecom': 'ClientTelecomID',
    'ClientUpdatePDSCache': None,

    # Con*: Contracts
    'Contract': 'ContractNumber',
    'ConAdHocAwaitingApproval': 'SequenceNo',
    'ConClientInitialBedRate': None,
    'ConClinicHistory': 'SequenceNo',
    'ConLeaveDiscountHistory': 'SequenceNo',

    # Not documented, but looks like Core
    'Deceased': None,  # or possibly TrustWideID (or just ClientID!)

    'DemClientDeletedDetails': None,

    # EP: E-Prescribing
    # ... with DA: Drug Administration
    # ... with DS: Drug Service
    'EPClientConditions': 'RowID',
    'EPClientPrescription': 'PrescriptionID',
    'EPClientSensitivities': None,  # UNVERIFIED: None? Joint PK on ProdID?
    'EPDiscretionaryDrugClientLink': None,
    'EPVariableDosageDrugLink': 'HistoryID',  # UNVERIFIED
    'EPClientAllergies': 'ReactionID',
    'DAConcurrencyControl': None,
    'DAIPPrescription': 'PrescriptionID',
    'DSBatchPatientGroups': None,
    'DSMedicationBatchContinue': None,
    'DSMedicationBatchLink': None,

    # Ims*: Inpatient Management System
    'ImsEventLeave': 'UniqueSequenceID',  # SequenceID
    'ImsEventMovement': None,
    'ImsEventRefno': None,  # Not in docs but looks like Core.
    'ImsEventRefnoBAKUP': None,  # [Sic.] Not in docs but looks like Core.

    # LR*: Legitimate Relationships
    'LRIdentifiedCache': None,

    # Mes*: messaging
    'MesLettersGenerated': 'Reference',

    # Mnt*: Mental Health module (re MHA detention)
    'MntArtAttendee': None,  # SequenceID being "of person within a meeting"
    'MntArtOutcome': None,  # ditto
    'MntArtPanel': None,  # ditto
    'MntArtRpts': None,  # ditto
    'MntArtRptsReceived': None,  # ditto
    'MntClientEctSection62': None,
    'MntClientMedSection62': None,
    'MntClientSectionDetailCareCoOrdinator': None,
    'MntClientSectionDetailCourtAppearance': None,
    'MntClientSectionDetailFMR': None,
    'MntClientSectionReview': None,

    # NDTMS*: Nation(al?) Drug Treatment Monitoring System

    # SNOMED*: SNOMED
    'SNOMED_Client': 'SC_ID',

    # UserAssess*: user assessment (= non-core?) tables.
    # See other default PK below: type12:

    # -------------------------------------------------------------------------
    # Non-core? No docs available.
    # -------------------------------------------------------------------------
    # Chd*: presumably, child development
    'ChdClientDevCheckBreastFeeding': None,
    # ... guess; DevChkSeqID is probably FK to ChdClientDevCheck.SequenceID

    # ??? But it has q1-q30, qu2-14, home, sch, comm... assessment tool...
    'CYPcurrentviewImport': None,  # not TrustWideID (which is non-unique)

    'GoldmineIfcMapping': None,  # no idea, really, and no data to explore

    'KP90ErrorLog': None,

    'ReportsOutpatientWaitersHashNotSeenReferrals': None,
    'ReportsOutpatientWaitersNotSeenReferrals': None,

    'UserAssesstfkcsa_childprev': 'type12_RowID',  # Keeping Children Safe Assessment subtable  # noqa
    'UserAssesstfkcsa_childs': 'type12_RowID',  # Keeping Children Safe Assessment subtable  # noqa
}

"""

RIO_6_2_ATYPICAL_PATIENT_ID_COLS = {
    'SNOMED_Client': 'SC_ClientID',
}



#!/usr/bin/env python
# crate_anon/crateweb/consent/migrations/0001_initial.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings
import crate_anon.crateweb.consent.models as consent_models
import crate_anon.crateweb.consent.storage as consent_storage
import crate_anon.crateweb.extra.fields as extra_fields

# !!! warning !!! some fields hard-code a local file path in /home/rudolf/...
# ... edited; no default is OK here; see
# https://docs.djangoproject.com/en/1.9/ref/files/storage/#the-filesystemstorage-class  # noqa


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CharityPaymentRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='When created')),  # noqa
                ('payee', models.CharField(max_length=255)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=8)),  # noqa
            ],
        ),
        migrations.CreateModel(
            name='ClinicianResponse',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='When created')),  # noqa
                ('token', models.CharField(max_length=20)),
                ('responded', models.BooleanField(default=False, verbose_name='Responded?')),  # noqa
                ('responded_at', models.DateTimeField(null=True, verbose_name='When responded')),  # noqa
                ('response_route', models.CharField(choices=[('e', 'E-mail'), ('w', 'Web')], max_length=1)),  # noqa
                ('email_choice', models.CharField(choices=[('y', 'Yes'), ('n', 'No'), ('more', 'Tell me more')], max_length=4)),  # noqa
                ('response', models.CharField(choices=[('R', 'R: Clinician asks RDBM to pass request to patient'), ('A', 'A: Clinician will pass the request to the patient'), ('B', 'B: Clinician vetoes on clinical grounds'), ('C', 'C: Patient is definitely ineligible'), ('D', 'D: Patient is dead/discharged or details are defunct')], max_length=1)),  # noqa
                ('veto_reason', models.TextField(blank=True, verbose_name='Reason for clinical veto')),  # noqa
                ('ineligible_reason', models.TextField(blank=True, verbose_name='Reason patient is ineligible')),  # noqa
                ('pt_uncontactable_reason', models.TextField(blank=True, verbose_name='Reason patient is not contactable')),  # noqa
                ('clinician_confirm_name', models.CharField(verbose_name='Type your name to confirm', max_length=255)),  # noqa
                ('charity_amount_due', models.DecimalField(default=0, decimal_places=2, max_digits=8)),  # noqa
            ],
        ),
        migrations.CreateModel(
            name='ConsentMode',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('decision_signed_by_patient', models.BooleanField(default=False, verbose_name='Request signed by patient?')),  # noqa
                ('decision_otherwise_directly_authorized_by_patient', models.BooleanField(default=False, verbose_name='Request otherwise directly authorized by patient?')),  # noqa
                ('decision_under16_signed_by_parent', models.BooleanField(default=False, verbose_name='Patient under 16 and request countersigned by parent?')),  # noqa
                ('decision_under16_signed_by_clinician', models.BooleanField(default=False, verbose_name='Patient under 16 and request countersigned by clinician?')),  # noqa
                ('decision_lack_capacity_signed_by_representative', models.BooleanField(default=False, verbose_name='Patient lacked capacity and request signed by authorized representative?')),  # noqa
                ('decision_lack_capacity_signed_by_clinician', models.BooleanField(default=False, verbose_name='Patient lacked capacity and request countersigned by clinician?')),  # noqa
                ('nhs_number', models.BigIntegerField(verbose_name='NHS number')),  # noqa
                ('current', models.BooleanField(default=False)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='When was this record created?')),  # noqa
                ('exclude_entirely', models.BooleanField(default=False, verbose_name='Exclude patient from Research Database entirely?')),  # noqa
                ('consent_mode', models.CharField(default='', choices=[('red', 'red'), ('yellow', 'yellow'), ('green', 'green')], verbose_name="Consent mode ('red', 'yellow', 'green')", max_length=10)),  # noqa
                ('consent_after_discharge', models.BooleanField(default=False, verbose_name='Consent given to contact patient after discharge?')),  # noqa
                ('max_approaches_per_year', models.PositiveSmallIntegerField(default=0, verbose_name='Maximum number of approaches permissible per year (0 = no limit)')),  # noqa
                ('other_requests', models.TextField(blank=True, verbose_name='Other special requests by patient')),  # noqa
                ('prefers_email', models.BooleanField(default=False, verbose_name='Patient prefers e-mail contact?')),  # noqa
                ('changed_by_clinician_override', models.BooleanField(default=False, verbose_name="Consent mode changed by clinician's override?")),  # noqa
                ('created_by', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ContactRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='When created')),  # noqa
                ('request_direct_approach', models.BooleanField(verbose_name='Request direct contact with patient if available (not contact with clinician first)')),  # noqa
                ('lookup_nhs_number', models.BigIntegerField(null=True, verbose_name='NHS number used for lookup')),  # noqa
                ('lookup_rid', models.CharField(verbose_name='Research ID used for lookup', null=True, max_length=128)),  # noqa
                ('lookup_mrid', models.CharField(verbose_name='Master research ID used for lookup', null=True, max_length=128)),  # noqa
                ('processed', models.BooleanField(default=False)),
                ('nhs_number', models.BigIntegerField(null=True, verbose_name='NHS number')),  # noqa
                ('approaches_in_past_year', models.PositiveIntegerField(null=True)),  # noqa
                ('decisions', models.TextField(blank=True, verbose_name='Decisions made')),  # noqa
                ('decided_no_action', models.BooleanField(default=False)),
                ('decided_send_to_researcher', models.BooleanField(default=False)),  # noqa
                ('decided_send_to_clinician', models.BooleanField(default=False)),  # noqa
                ('clinician_involvement', models.PositiveSmallIntegerField(choices=[(0, 'No clinician involvement required or requested'), (1, 'Clinician involvement requested by researchers'), (2, 'Clinician involvement required by YELLOW consent mode'), (3, 'Clinician involvement required by UNKNOWN consent mode')], null=True)),  # noqa
                ('consent_withdrawn', models.BooleanField(default=False)),
                ('consent_withdrawn_at', models.DateTimeField(null=True, verbose_name='When consent withdrawn')),  # noqa
                ('consent_mode', models.ForeignKey(to='consent.ConsentMode', null=True)),  # noqa
            ],
        ),
        migrations.CreateModel(
            name='DummyPatientSourceInfo',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('pt_local_id_description', models.CharField(blank=True, verbose_name='Description of database-specific ID', max_length=100)),  # noqa
                ('pt_local_id_number', models.BigIntegerField(blank=True, null=True, verbose_name='Database-specific ID')),  # noqa
                ('pt_dob', models.DateField(blank=True, null=True, verbose_name='Patient date of birth')),  # noqa
                ('pt_dod', models.DateField(blank=True, null=True, verbose_name='Patient date of death (NULL if alive)')),  # noqa
                ('pt_dead', models.BooleanField(default=False, verbose_name='Patient is dead')),  # noqa
                ('pt_discharged', models.NullBooleanField(verbose_name='Patient discharged')),  # noqa
                ('pt_sex', models.CharField(choices=[('M', 'Male'), ('F', 'Female'), ('X', 'Inderminate/intersex'), ('?', 'Unknown')], blank=True, verbose_name='Patient sex', max_length=1)),  # noqa
                ('pt_title', models.CharField(blank=True, verbose_name='Patient title', max_length=20)),  # noqa
                ('pt_first_name', models.CharField(blank=True, verbose_name='Patient first name', max_length=100)),  # noqa
                ('pt_last_name', models.CharField(blank=True, verbose_name='Patient last name', max_length=100)),  # noqa
                ('pt_address_1', models.CharField(blank=True, verbose_name='Patient address line 1', max_length=100)),  # noqa
                ('pt_address_2', models.CharField(blank=True, verbose_name='Patient address line 2', max_length=100)),  # noqa
                ('pt_address_3', models.CharField(blank=True, verbose_name='Patient address line 3', max_length=100)),  # noqa
                ('pt_address_4', models.CharField(blank=True, verbose_name='Patient address line 4', max_length=100)),  # noqa
                ('pt_address_5', models.CharField(blank=True, verbose_name='Patient address line 5 (county)', max_length=100)),  # noqa
                ('pt_address_6', models.CharField(blank=True, verbose_name='Patient address line 6 (postcode)', max_length=100)),  # noqa
                ('pt_address_7', models.CharField(blank=True, verbose_name='Patient address line 7 (country)', max_length=100)),  # noqa
                ('pt_telephone', models.CharField(blank=True, verbose_name='Patient telephone', max_length=20)),  # noqa
                ('pt_email', models.EmailField(blank=True, verbose_name='Patient email', max_length=254)),  # noqa
                ('gp_title', models.CharField(blank=True, verbose_name='GP title', max_length=20)),  # noqa
                ('gp_first_name', models.CharField(blank=True, verbose_name='GP first name', max_length=100)),  # noqa
                ('gp_last_name', models.CharField(blank=True, verbose_name='GP last name', max_length=100)),  # noqa
                ('gp_address_1', models.CharField(blank=True, verbose_name='GP address line 1', max_length=100)),  # noqa
                ('gp_address_2', models.CharField(blank=True, verbose_name='GP address line 2', max_length=100)),  # noqa
                ('gp_address_3', models.CharField(blank=True, verbose_name='GP address line 3', max_length=100)),  # noqa
                ('gp_address_4', models.CharField(blank=True, verbose_name='GP address line 4', max_length=100)),  # noqa
                ('gp_address_5', models.CharField(blank=True, verbose_name='GP address line 5 (county)', max_length=100)),  # noqa
                ('gp_address_6', models.CharField(blank=True, verbose_name='GP address line 6 (postcode)', max_length=100)),  # noqa
                ('gp_address_7', models.CharField(blank=True, verbose_name='GP address line 7 (country)', max_length=100)),  # noqa
                ('gp_telephone', models.CharField(blank=True, verbose_name='GP telephone', max_length=20)),  # noqa
                ('gp_email', models.EmailField(blank=True, verbose_name='GP email', max_length=254)),  # noqa
                ('clinician_title', models.CharField(blank=True, verbose_name='Clinician title', max_length=20)),  # noqa
                ('clinician_first_name', models.CharField(blank=True, verbose_name='Clinician first name', max_length=100)),  # noqa
                ('clinician_last_name', models.CharField(blank=True, verbose_name='Clinician last name', max_length=100)),  # noqa
                ('clinician_address_1', models.CharField(blank=True, verbose_name='Clinician address line 1', max_length=100)),  # noqa
                ('clinician_address_2', models.CharField(blank=True, verbose_name='Clinician address line 2', max_length=100)),  # noqa
                ('clinician_address_3', models.CharField(blank=True, verbose_name='Clinician address line 3', max_length=100)),  # noqa
                ('clinician_address_4', models.CharField(blank=True, verbose_name='Clinician address line 4', max_length=100)),  # noqa
                ('clinician_address_5', models.CharField(blank=True, verbose_name='Clinician address line 5 (county)', max_length=100)),  # noqa
                ('clinician_address_6', models.CharField(blank=True, verbose_name='Clinician address line 6 (postcode)', max_length=100)),  # noqa
                ('clinician_address_7', models.CharField(blank=True, verbose_name='Clinician address line 7 (country)', max_length=100)),  # noqa
                ('clinician_telephone', models.CharField(blank=True, verbose_name='Clinician telephone', max_length=20)),  # noqa
                ('clinician_email', models.EmailField(blank=True, verbose_name='Clinician email', max_length=254)),  # noqa
                ('clinician_is_consultant', models.BooleanField(default=False, verbose_name='Clinician is a consultant')),  # noqa
                ('clinician_signatory_title', models.CharField(blank=True, verbose_name="Clinician's title for signature (e.g. 'Consultant psychiatrist')", max_length=100)),  # noqa
                ('nhs_number', models.BigIntegerField(unique=True, verbose_name='NHS number')),  # noqa
            ],
            options={
                'verbose_name_plural': 'Dummy patient source information',
            },
        ),
        migrations.CreateModel(
            name='Email',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='When created')),  # noqa
                ('sender', models.CharField(default='CPFT Research Database - DO NOT REPLY <noreply@cpft.nhs.uk>', max_length=255)),  # noqa
                ('recipient', models.CharField(max_length=255)),
                ('subject', models.CharField(max_length=255)),
                ('msg_text', models.TextField()),
                ('msg_html', models.TextField()),
                ('to_clinician', models.BooleanField(default=False)),
                ('to_researcher', models.BooleanField(default=False)),
                ('to_patient', models.BooleanField(default=False)),
                ('contact_request', models.ForeignKey(to='consent.ContactRequest', null=True)),  # noqa
            ],
        ),
        migrations.CreateModel(
            name='EmailAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('file', models.FileField(upload_to='', storage=consent_storage.CustomFileSystemStorage(base_url='download_privatestorage'))),  # noqa
                ('sent_filename', models.CharField(null=True, max_length=255)),
                ('content_type', models.CharField(null=True, max_length=255)),
                ('owns_file', models.BooleanField(default=False)),
                ('email', models.ForeignKey(to='consent.Email')),
            ],
        ),
        migrations.CreateModel(
            name='EmailTransmission',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('at', models.DateTimeField(auto_now_add=True, verbose_name='When sent')),  # noqa
                ('sent', models.BooleanField(default=False)),
                ('failure_reason', models.TextField(verbose_name='Reason sending failed')),  # noqa
                ('by', models.ForeignKey(to=settings.AUTH_USER_MODEL, related_name='emailtransmissions', null=True)),  # noqa
                ('email', models.ForeignKey(to='consent.Email')),
            ],
        ),
        migrations.CreateModel(
            name='Leaflet',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('name', models.CharField(choices=[('cpft_tpir', 'CPFT: Taking part in research'), ('nihr_yhrsl', 'NIHR: Your health records save lives'), ('cpft_trafficlight_choice', 'CPFT: traffic-light choice'), ('cpft_clinres', 'CPFT: clinical research')], verbose_name='leaflet name', unique=True, max_length=50)),  # noqa
                ('pdf', extra_fields.ContentTypeRestrictedFileField(upload_to=consent_models.leaflet_upload_to, blank=True, storage=consent_storage.CustomFileSystemStorage(base_url='download_privatestorage'))),  # noqa
            ],
        ),
        migrations.CreateModel(
            name='Letter',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='When created')),  # noqa
                ('pdf', models.FileField(upload_to='', storage=consent_storage.CustomFileSystemStorage(base_url='download_privatestorage'))),  # noqa
                ('to_clinician', models.BooleanField(default=False)),
                ('to_researcher', models.BooleanField(default=False)),
                ('to_patient', models.BooleanField(default=False)),
                ('rdbm_may_view', models.BooleanField(default=False)),
                ('sent_manually_at', models.DateTimeField(null=True)),
                ('contact_request', models.ForeignKey(to='consent.ContactRequest', null=True)),  # noqa
            ],
        ),
        migrations.CreateModel(
            name='PatientLookup',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('pt_local_id_description', models.CharField(blank=True, verbose_name='Description of database-specific ID', max_length=100)),  # noqa
                ('pt_local_id_number', models.BigIntegerField(blank=True, null=True, verbose_name='Database-specific ID')),  # noqa
                ('pt_dob', models.DateField(blank=True, null=True, verbose_name='Patient date of birth')),  # noqa
                ('pt_dod', models.DateField(blank=True, null=True, verbose_name='Patient date of death (NULL if alive)')),  # noqa
                ('pt_dead', models.BooleanField(default=False, verbose_name='Patient is dead')),  # noqa
                ('pt_discharged', models.NullBooleanField(verbose_name='Patient discharged')),  # noqa
                ('pt_sex', models.CharField(choices=[('M', 'Male'), ('F', 'Female'), ('X', 'Inderminate/intersex'), ('?', 'Unknown')], blank=True, verbose_name='Patient sex', max_length=1)),  # noqa
                ('pt_title', models.CharField(blank=True, verbose_name='Patient title', max_length=20)),  # noqa
                ('pt_first_name', models.CharField(blank=True, verbose_name='Patient first name', max_length=100)),  # noqa
                ('pt_last_name', models.CharField(blank=True, verbose_name='Patient last name', max_length=100)),  # noqa
                ('pt_address_1', models.CharField(blank=True, verbose_name='Patient address line 1', max_length=100)),  # noqa
                ('pt_address_2', models.CharField(blank=True, verbose_name='Patient address line 2', max_length=100)),  # noqa
                ('pt_address_3', models.CharField(blank=True, verbose_name='Patient address line 3', max_length=100)),  # noqa
                ('pt_address_4', models.CharField(blank=True, verbose_name='Patient address line 4', max_length=100)),  # noqa
                ('pt_address_5', models.CharField(blank=True, verbose_name='Patient address line 5 (county)', max_length=100)),  # noqa
                ('pt_address_6', models.CharField(blank=True, verbose_name='Patient address line 6 (postcode)', max_length=100)),  # noqa
                ('pt_address_7', models.CharField(blank=True, verbose_name='Patient address line 7 (country)', max_length=100)),  # noqa
                ('pt_telephone', models.CharField(blank=True, verbose_name='Patient telephone', max_length=20)),  # noqa
                ('pt_email', models.EmailField(blank=True, verbose_name='Patient email', max_length=254)),  # noqa
                ('gp_title', models.CharField(blank=True, verbose_name='GP title', max_length=20)),  # noqa
                ('gp_first_name', models.CharField(blank=True, verbose_name='GP first name', max_length=100)),  # noqa
                ('gp_last_name', models.CharField(blank=True, verbose_name='GP last name', max_length=100)),  # noqa
                ('gp_address_1', models.CharField(blank=True, verbose_name='GP address line 1', max_length=100)),  # noqa
                ('gp_address_2', models.CharField(blank=True, verbose_name='GP address line 2', max_length=100)),  # noqa
                ('gp_address_3', models.CharField(blank=True, verbose_name='GP address line 3', max_length=100)),  # noqa
                ('gp_address_4', models.CharField(blank=True, verbose_name='GP address line 4', max_length=100)),  # noqa
                ('gp_address_5', models.CharField(blank=True, verbose_name='GP address line 5 (county)', max_length=100)),  # noqa
                ('gp_address_6', models.CharField(blank=True, verbose_name='GP address line 6 (postcode)', max_length=100)),  # noqa
                ('gp_address_7', models.CharField(blank=True, verbose_name='GP address line 7 (country)', max_length=100)),  # noqa
                ('gp_telephone', models.CharField(blank=True, verbose_name='GP telephone', max_length=20)),  # noqa
                ('gp_email', models.EmailField(blank=True, verbose_name='GP email', max_length=254)),  # noqa
                ('clinician_title', models.CharField(blank=True, verbose_name='Clinician title', max_length=20)),  # noqa
                ('clinician_first_name', models.CharField(blank=True, verbose_name='Clinician first name', max_length=100)),  # noqa
                ('clinician_last_name', models.CharField(blank=True, verbose_name='Clinician last name', max_length=100)),  # noqa
                ('clinician_address_1', models.CharField(blank=True, verbose_name='Clinician address line 1', max_length=100)),  # noqa
                ('clinician_address_2', models.CharField(blank=True, verbose_name='Clinician address line 2', max_length=100)),  # noqa
                ('clinician_address_3', models.CharField(blank=True, verbose_name='Clinician address line 3', max_length=100)),  # noqa
                ('clinician_address_4', models.CharField(blank=True, verbose_name='Clinician address line 4', max_length=100)),  # noqa
                ('clinician_address_5', models.CharField(blank=True, verbose_name='Clinician address line 5 (county)', max_length=100)),  # noqa
                ('clinician_address_6', models.CharField(blank=True, verbose_name='Clinician address line 6 (postcode)', max_length=100)),  # noqa
                ('clinician_address_7', models.CharField(blank=True, verbose_name='Clinician address line 7 (country)', max_length=100)),  # noqa
                ('clinician_telephone', models.CharField(blank=True, verbose_name='Clinician telephone', max_length=20)),  # noqa
                ('clinician_email', models.EmailField(blank=True, verbose_name='Clinician email', max_length=254)),  # noqa
                ('clinician_is_consultant', models.BooleanField(default=False, verbose_name='Clinician is a consultant')),  # noqa
                ('clinician_signatory_title', models.CharField(blank=True, verbose_name="Clinician's title for signature (e.g. 'Consultant psychiatrist')", max_length=100)),  # noqa
                ('nhs_number', models.BigIntegerField(verbose_name='NHS number used for lookup')),  # noqa
                ('lookup_at', models.DateTimeField(auto_now_add=True, verbose_name='When fetched from clinical database')),  # noqa
                ('source_db', models.CharField(choices=[('dummy_clinical', 'Dummy clinical database for testing'), ('cpft_crs', 'CPFT Care Records System (CRS) 2005-2012'), ('cpft_rio', 'CPFT RiO 2013-')], verbose_name='Source database used for lookup', max_length=20)),  # noqa
                ('decisions', models.TextField(blank=True, verbose_name='Decisions made during lookup')),  # noqa
                ('secret_decisions', models.TextField(blank=True, verbose_name='Secret (identifying) decisions made during lookup')),  # noqa
                ('pt_found', models.BooleanField(default=False, verbose_name='Patient found')),  # noqa
                ('gp_found', models.BooleanField(default=False, verbose_name='GP found')),  # noqa
                ('clinician_found', models.BooleanField(default=False, verbose_name='Clinician found')),  # noqa
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PatientResponse',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('decision_signed_by_patient', models.BooleanField(default=False, verbose_name='Request signed by patient?')),  # noqa
                ('decision_otherwise_directly_authorized_by_patient', models.BooleanField(default=False, verbose_name='Request otherwise directly authorized by patient?')),  # noqa
                ('decision_under16_signed_by_parent', models.BooleanField(default=False, verbose_name='Patient under 16 and request countersigned by parent?')),  # noqa
                ('decision_under16_signed_by_clinician', models.BooleanField(default=False, verbose_name='Patient under 16 and request countersigned by clinician?')),  # noqa
                ('decision_lack_capacity_signed_by_representative', models.BooleanField(default=False, verbose_name='Patient lacked capacity and request signed by authorized representative?')),  # noqa
                ('decision_lack_capacity_signed_by_clinician', models.BooleanField(default=False, verbose_name='Patient lacked capacity and request countersigned by clinician?')),  # noqa
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='When created')),  # noqa
                ('response', models.PositiveSmallIntegerField(choices=[(1, '1: Yes'), (2, '2: No')], null=True, verbose_name="Patient's response")),  # noqa
                ('contact_request', models.OneToOneField(to='consent.ContactRequest', related_name='patient_response')),  # noqa
                ('recorded_by', models.ForeignKey(to=settings.AUTH_USER_MODEL, null=True)),  # noqa
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Study',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('institutional_id', models.PositiveIntegerField(unique=True, verbose_name='Institutional (e.g. NHS Trust) study number')),  # noqa
                ('title', models.CharField(verbose_name='Study title', max_length=255)),  # noqa
                ('registered_at', models.DateTimeField(blank=True, null=True, verbose_name='When was the study registered?')),  # noqa
                ('summary', models.TextField(verbose_name='Summary of study')),
                ('search_methods_planned', models.TextField(blank=True, verbose_name='Search methods planned')),  # noqa
                ('patient_contact', models.BooleanField(verbose_name='Involves patient contact?')),  # noqa
                ('include_under_16s', models.BooleanField(verbose_name='Include patients under 16?')),  # noqa
                ('include_lack_capacity', models.BooleanField(verbose_name='Include patients lacking capacity?')),  # noqa
                ('clinical_trial', models.BooleanField(verbose_name='Clinical trial (CTIMP)?')),  # noqa
                ('include_discharged', models.BooleanField(verbose_name='Include discharged patients?')),  # noqa
                ('request_direct_approach', models.BooleanField(verbose_name='Researchers request direct approach to patients?')),  # noqa
                ('approved_by_rec', models.BooleanField(verbose_name='Approved by REC?')),  # noqa
                ('rec_reference', models.CharField(blank=True, verbose_name='Research Ethics Committee reference', max_length=50)),  # noqa
                ('approved_locally', models.BooleanField(verbose_name='Approved by local institution?')),  # noqa
                ('local_approval_at', models.DateTimeField(blank=True, null=True, verbose_name='When approved by local institution?')),  # noqa
                ('study_details_pdf', extra_fields.ContentTypeRestrictedFileField(upload_to=consent_models.study_details_upload_to, blank=True, storage=consent_storage.CustomFileSystemStorage(base_url='download_privatestorage'))),  # noqa
                ('subject_form_template_pdf', extra_fields.ContentTypeRestrictedFileField(upload_to=consent_models.study_form_upload_to, blank=True, storage=consent_storage.CustomFileSystemStorage(base_url='download_privatestorage'))),  # noqa
                ('lead_researcher', models.ForeignKey(to=settings.AUTH_USER_MODEL, related_name='studies_as_lead')),  # noqa
                ('researchers', models.ManyToManyField(to=settings.AUTH_USER_MODEL, blank=True, related_name='studies_as_researcher')),  # noqa
            ],
            options={
                'verbose_name_plural': 'studies',
            },
        ),
        migrations.CreateModel(
            name='TeamRep',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),  # noqa
                ('team', models.CharField(choices=[('dummy_team_one', 'dummy_team_one'), ('dummy_team_two', 'dummy_team_two'), ('dummy_team_three', 'dummy_team_three')], verbose_name='Team description', unique=True, max_length=100)),  # noqa
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name_plural': 'clinical team representatives',
                'verbose_name': 'clinical team representative',
            },
        ),
        migrations.AddField(
            model_name='letter',
            name='study',
            field=models.ForeignKey(to='consent.Study', null=True),
        ),
        migrations.AddField(
            model_name='email',
            name='letter',
            field=models.ForeignKey(to='consent.Letter', null=True),
        ),
        migrations.AddField(
            model_name='email',
            name='study',
            field=models.ForeignKey(to='consent.Study', null=True),
        ),
        migrations.AddField(
            model_name='contactrequest',
            name='patient_lookup',
            field=models.ForeignKey(to='consent.PatientLookup', null=True),
        ),
        migrations.AddField(
            model_name='contactrequest',
            name='request_by',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='contactrequest',
            name='study',
            field=models.ForeignKey(to='consent.Study'),
        ),
        migrations.AddField(
            model_name='clinicianresponse',
            name='contact_request',
            field=models.OneToOneField(to='consent.ContactRequest', related_name='clinician_response'),  # noqa
        ),
    ]



#!/usr/bin/env python
# crate_anon/nlp_manager/parse_cognitive.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
from typing import Optional

from crate_anon.nlp_manager.nlp_definition import NlpDefinition
from crate_anon.nlp_manager.regex_numbers import UNSIGNED_INTEGER
from crate_anon.nlp_manager.regex_parser import (
    APOSTROPHE,
    NumeratorOutOfDenominatorParser,
    ValidatorBase,
    WORD_BOUNDARY,
)
from crate_anon.nlp_manager.regex_units import OUT_OF_SEPARATOR

log = logging.getLogger(__name__)


# =============================================================================
# Mini-mental state examination (MMSE)
# =============================================================================

class Mmse(NumeratorOutOfDenominatorParser):
    """Mini-mental state examination (MMSE)."""
    MMSE = r"""
        (?: {WORD_BOUNDARY}
            (?: MMSE | mini[-\s]*mental (?: \s+ state)?
                       (?: \s+ exam(?:ination)? )? )
        {WORD_BOUNDARY} )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    NAME = "MMSE"

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            commit=commit,
            variable_name=self.NAME,
            variable_regex_str=self.MMSE,
            expected_denominator=30,
            take_absolute=True
        )

    def test(self, verbose: bool = False) -> None:
        self.test_numerator_denominator_parser([
            ("MMSE", []),  # should fail; no values
            ("MMSE 30/30", [(30, 30)]),
            ("MMSE 25 / 30", [(25, 30)]),
            ("MMSE 25 / 29", [(25, 29)]),
            ("MMSE 25 / 31", [(25, 31)]),
            ("mini-mental state exam 30", [(30, None)]),
            ("minimental 25", [(25, None)]),
            ("MMSE 30", [(30, None)]),
            ("MMSE-27", [(27, None)]),
            ("MMSE score was 30", [(30, None)]),
            ("ACE 79", []),
        ], verbose=verbose)


class MmseValidator(ValidatorBase):
    """Validator for Mmse (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Mmse.MMSE],
                         validated_variable=Mmse.NAME,
                         commit=commit)


# =============================================================================
# Addenbrooke's Cognitive Examination (ACE, ACE-R, ACE-III)
# =============================================================================

class Ace(NumeratorOutOfDenominatorParser):
    """Addenbrooke's Cognitive Examination (ACE, ACE-R, ACE-III)."""
    NAME = "ACE"
    ACE = r"""
        (?: {WORD_BOUNDARY}
            (?: ACE | (?: Addenbrooke{APOSTROPHE}?s \s+ cognitive \s+
                          (?: (?:evaluation) | exam(?:ination)? ) ) )
            (?: \s* -? \s*
                (?: R | III | 111
                    # or: 3 when not followed by an "out of X" expression
                    | (?: 3 (?! \s* {out_of} \s* {uint}))
                ) \b
            )?+
        {WORD_BOUNDARY} )
    """.format(
        WORD_BOUNDARY=WORD_BOUNDARY,
        APOSTROPHE=APOSTROPHE,
        out_of=OUT_OF_SEPARATOR,
        uint=UNSIGNED_INTEGER
    )
    # ... note the possessive "?+" above; see tests below.

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            commit=commit,
            variable_name=self.NAME,
            variable_regex_str=self.ACE,
            expected_denominator=100,
            take_absolute=True
        )

    def test(self, verbose: bool = False) -> None:
        self.test_numerator_denominator_parser([
            ("MMSE", []),
            ("MMSE 30/30", []),
            ("MMSE 25 / 30", []),
            ("mini-mental state exam 30", []),
            ("minimental 25", []),
            ("MMSE 30", []),
            ("ACE 79", [(79, None)]),
            ("ACE 79/100", [(79, 100)]),
            ("ACE 79/95", [(79, 95)]),
            ("ACE 79 / 100", [(79, 100)]),
            ("Addenbrooke's cognitive examination 79", [(79, None)]),
            ("Addenbrookes cognitive evaluation 79", [(79, None)]),
            ("ACE-R 79", [(79, None)]),
            ("ACE-R 79 out of 100", [(79, 100)]),
            ("ACE-III 79", [(79, None)]),
            ("ACE-III score was 79", [(79, None)]),
            ("ACE R 79", [(79, None)]),
            ("ACE III 79", [(79, None)]),
            ("ACE-82", [(82, None)]),
            ("ACE 111 99", [(99, None)]),  # "ACE 111" (for III) from real data
            # Note the difficulties created by the "ACE-3" representation of
            # the task's name. We have to get these right:
            ("ACE-3 79", [(79, None)]),
            ("ACE 3 79", [(79, None)]),
            ("ACE 3 79/100", [(79, 100)]),
            ("ACE 3 3", [(3, None)]),
            ("ACE 3 3/100", [(3, 100)]),
            # ... but also a score of 3 (!) on the older ACE:
            ("ACE 3/100", [(3, 100)]),
            ("ACE 3 out of 100", [(3, 100)]),
            # - This next one is ambiguous. Reference to new task? To old
            #   score? Making the "3" optional as part of the task name means
            #   that this will be accepted by the regex as a score.
            # - We need a special exception to get "ACE 3" not to give a score.
            # - We do this with a "possessive" quantifier on the "3" (or
            #   similar) part of the ACE descriptor.
            # - http://www.rexegg.com/regex-quantifiers.html
            # - Possessive quantifiers are in regex, not re:
            #   https://pypi.python.org/pypi/regex
            #   https://docs.python.org/3.5/library/re.html
            # - Ah, no. That makes "ACE 3/100" fail.
            # - But if we combine a possessive "3" with saying "3 unless it's
            #   "3 out of...", then we win.
            ("ACE 3", []),
            ("ACE 3/MOCA", []),
            ("ACE 3 / MOCA", []),
        ], verbose=verbose)


class AceValidator(ValidatorBase):
    """Validator for Ace (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Ace.ACE],
                         validated_variable=Ace.NAME,
                         commit=commit)

    def test(self, verbose: bool = False) -> None:
        self.test_validator([
            ("pass me my mace, my boy", False),
            ("he scored 10 on the ACE today", True),
            ("he scored 10 on the ACE 3 today", True),
            ("he scored 10 on the ACE3 today", True),
            ("ACE 3/100", True),
            ("ACE 3 3/100", True),
            ("ACE3 4", True),
            ("ACE 3", True),
            ("ACE3", True),
            ("ACE 3/MOCA", True),
            ("ACE 3 / MOCA", True),
        ], verbose=verbose)


# =============================================================================
# Mini-Addenbrooke's Cognitive Examination (M-ACE)
# =============================================================================

class MiniAce(NumeratorOutOfDenominatorParser):
    """Mini-Addenbrooke's Cognitive Examination (M-ACE)."""
    MACE = r"""
        (?: {WORD_BOUNDARY}
            (?: mini | M ) \s* -? \s*
            (?: ACE | (?: Addenbrooke{APOSTROPHE}?s \s+ cognitive \s+
                          (?: (?:evaluation) | exam(?:ination)? ) ) )
        {WORD_BOUNDARY} )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY, APOSTROPHE=APOSTROPHE)
    NAME = "MiniACE"

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            commit=commit,
            variable_name=self.NAME,
            variable_regex_str=self.MACE,
            expected_denominator=30,
            take_absolute=True
        )

    def test(self, verbose: bool = False) -> None:
        self.test_numerator_denominator_parser([
            ("MMSE 30", []),
            ("ACE 79", []),
            ("ACE 79/100", []),
            ("Addenbrooke's cognitive examination 79", []),
            ("Addenbrookes cognitive evaluation 79", []),
            ("mini-Addenbrooke's cognitive examination 79", [(79, None)]),
            ("mini-Addenbrooke’s cognitive examination 79", [(79, None)]),
            ("mini-Addenbrookes cognitive evaluation 79", [(79, None)]),
            ("M-ACE 20", [(20, None)]),
            ("M-ACE score is 20", [(20, None)]),
            ("M-ACE 29/30", [(29, 30)]),
            ("M-ACE 29/29", [(29, 29)]),
            ("MACE 29", [(29, None)]),
            ("MACE-29", [(29, None)]),
            ("mini-ACE 29", [(29, None)]),
        ], verbose=verbose)


class MiniAceValidator(ValidatorBase):
    """Validator for MiniAce (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[MiniAce.MACE],
                         validated_variable=MiniAce.NAME,
                         commit=commit)


# =============================================================================
# Montreal Cognitive Assessment (MOCA)
# =============================================================================

class Moca(NumeratorOutOfDenominatorParser):
    """Montreal Cognitive Assessment (MOCA)."""
    MOCA = r"""
        (?: {WORD_BOUNDARY}
            (?: MOCA | (?: Montreal \s+ cognitive \s+ assessment ) )
        {WORD_BOUNDARY} )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    NAME = "MOCA"

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            commit=commit,
            variable_name=self.NAME,
            variable_regex_str=self.MOCA,
            expected_denominator=30,
            take_absolute=True
        )

    def test(self, verbose: bool = False) -> None:
        self.test_numerator_denominator_parser([
            ("MOCA 30", [(30, None)]),
            ("MOCA 30/30", [(30, 30)]),
            ("MOCA 25/30", [(25, 30)]),
            ("MOCA score was 25", [(25, None)]),
            ("MOCA 25/29", [(25, 29)]),
            ("MOCA-25", [(25, None)]),
            ("Montreal Cognitive Assessment 25/30", [(25, 30)]),
        ])


class MocaValidator(ValidatorBase):
    """Validator for MiniAce (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[MiniAce.MACE],
                         validated_variable=MiniAce.NAME,
                         commit=commit)


# =============================================================================
#  Command-line entry point
# =============================================================================

def test_all(verbose: bool = False) -> None:
    mmse = Mmse(None, None)
    mmse.test(verbose=verbose)

    ace = Ace(None, None)
    ace.test(verbose=verbose)
    ace_validator = AceValidator(None, None)
    ace_validator.test(verbose=verbose)

    mace = MiniAce(None, None)
    mace.test(verbose=verbose)

    moca = Moca(None, None)
    moca.test(verbose=verbose)


if __name__ == '__main__':
    test_all(verbose=True)


# support also "scored X on the MOCA"?



#!/usr/bin/env python
# crate_anon/crateweb/research/models.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

See notes_on_pickle_json.txt

THE STANDARD PYTHON REPRESENTATION USED IS A DICTIONARY LIKE THIS:

    {
        __type__: 'MyClass',
        args: [some, positional, args],
        kwargs: {
            'some': 1,
            'named': 'hello',
            'args': [2, 3, 4],
        }
    }

We will call this an InitDict.
Sometimes positional arguments aren't necessary and it's convenient to work
also with the simpler dictionary:

    {
        'some': 1,
        'named': 'hello',
        'args': [2, 3, 4],
    }

... which we'll call a KwargsDict.

"""

import datetime
import json
import logging
import pprint
import sys
from typing import Any, Callable, Dict, List, TextIO, Tuple, Type

from django.core.exceptions import ValidationError
from django.db.models import TextField

from crate_anon.common.lang import auto_repr

Instance = Any
ClassType = Type[object]

InitDict = Dict[str, Any]
KwargsDict = Dict[str, Any]
ArgsList = List[Any]

ArgsKwargsTuple = Tuple[ArgsList, KwargsDict]

InstanceToDictFnType = Callable[[Instance], Dict]
DictToInstanceFnType = Callable[[Dict, ClassType], Instance]
DefaultFactoryFnType = Callable[[], Instance]
InitArgsKwargsFnType = Callable[[Instance], ArgsKwargsTuple]
InitKwargsFnType = Callable[[Instance], KwargsDict]
InstanceToInitDictFnType = Callable[[Instance], InitDict]

log = logging.getLogger(__name__)

# =============================================================================
# Constants for external use
# =============================================================================

METHOD_NO_ARGS = 'no_args'
METHOD_SIMPLE = 'simple'
METHOD_STRIP_UNDERSCORE = 'strip_underscore'
METHOD_PROVIDES_INIT_ARGS_KWARGS = 'provides_init_args_kwargs'
METHOD_PROVIDES_INIT_KWARGS = 'provides_init_kwargs'

# =============================================================================
# Constants for internal use
# =============================================================================

DEBUG = False

ARGS_LABEL = 'args'
KWARGS_LABEL = 'kwargs'
TYPE_LABEL = '__type__'

INIT_ARGS_KWARGS_FN_NAME = 'init_args_kwargs'
INIT_KWARGS_FN_NAME = 'init_kwargs'


# =============================================================================
# Simple dictionary manipulation
# =============================================================================

def args_kwargs_to_initdict(args: ArgsList, kwargs: KwargsDict) -> InitDict:
    return {ARGS_LABEL: args,
            KWARGS_LABEL: kwargs}


def kwargs_to_initdict(kwargs: KwargsDict) -> InitDict:
    return {ARGS_LABEL: [],
            KWARGS_LABEL: kwargs}


# noinspection PyUnusedLocal
def obj_with_no_args_to_init_dict(obj: Any) -> InitDict:
    return {ARGS_LABEL: [],
            KWARGS_LABEL: {}}


def strip_leading_underscores_from_keys(d: Dict) -> Dict:
    newdict = {}
    for k, v in d.items():
        if k.startswith('_'):
            k = k[1:]
            if k in newdict:
                raise ValueError("Attribute conflict: _{k}, {k}".format(k=k))
        newdict[k] = v
    return newdict


def verify_initdict(initdict: InitDict) -> None:
    if (not isinstance(initdict, dict) or
            ARGS_LABEL not in initdict or
            KWARGS_LABEL not in initdict):
        raise ValueError("Not an InitDict dictionary")


# =============================================================================
# InitDict -> class instance
# =============================================================================

def initdict_to_instance(d: InitDict, cls: ClassType) -> Any:
    """
    Converse of simple_to_dict().
    Given that JSON dictionary, we will end up re-instantiating the class with

        d = {'a': 1, 'b': 2, 'c': 3}
        new_x = SimpleClass(**d)

    We'll also support arbitrary creation, by using both *args and **kwargs.
    """
    args = d.get(ARGS_LABEL, [])
    kwargs = d.get(KWARGS_LABEL, {})
    return cls(*args, **kwargs)


# =============================================================================
# Class instance -> InitDict, in various ways
# =============================================================================

def instance_to_initdict_simple(obj: Any) -> InitDict:
    """
    For use when object attributes (found in obj.__dict__) should be mapped
    directly to the serialized JSON dictionary. Typically used for classes
    like:

class SimpleClass(object):
    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    Here, after

        x = SimpleClass(a=1, b=2, c=3)

    we will find that

        x.__dict__ == {'a': 1, 'b': 2, 'c': 3}

    and that dictionary is a reasonable thing to serialize to JSON as keyword
    arguments.

    We'll also support arbitrary creation, by using both *args and **kwargs.
    We may not use this format much, but it has the advantage of being an
    arbitrarily correct format for Python class construction.
    """
    return kwargs_to_initdict(obj.__dict__)


def instance_to_initdict_stripping_underscores(obj: Instance) -> InitDict:
    """
    This is appropriate when a class uses a '_' prefix for all its __init__
    parameters, like this:

class UnderscoreClass(object):
    def __init__(self, a, b, c):
        self._a = a
        self._b = b
        self._c = c

    Here, after

        y = UnderscoreClass(a=1, b=2, c=3)

    we will find that

        y.__dict__ == {'_a': 1, '_b': 2, '_c': 3}

    but we would like to serialize the parameters we can pass back to __init__,
    by removing the leading underscores, like this:

        {'a': 1, 'b': 2, 'c': 3}
    """
    return kwargs_to_initdict(
        strip_leading_underscores_from_keys(obj.__dict__))


def wrap_kwargs_to_initdict(init_kwargs_fn: InitKwargsFnType,
                            typename: str,
                            check_result: bool = True) \
        -> InstanceToInitDictFnType:
    """
    Wraps a function producing a KwargsDict, making it into a function
    producing an InitDict.
    """
    def wrapper(obj: Instance) -> InitDict:
        result = init_kwargs_fn(obj)
        if check_result:
            if not isinstance(result, dict):
                raise ValueError(
                    "Class {} failed to provide a kwargs dict and "
                    "provided instead: {}".format(typename, repr(result)))
        return kwargs_to_initdict(init_kwargs_fn(obj))

    return wrapper


def wrap_args_kwargs_to_initdict(init_args_kwargs_fn: InitArgsKwargsFnType,
                                 typename: str,
                                 check_result: bool = True) \
        -> InstanceToInitDictFnType:
    """
    Wraps a function producing a KwargsDict, making it into a function
    producing an InitDict.
    """
    def wrapper(obj: Instance) -> InitDict:
        result = init_args_kwargs_fn(obj)
        if check_result:
            if (not isinstance(result, tuple) or
                    not len(result) == 2 or
                    not isinstance(result[0], list) or
                    not isinstance(result[1], dict)):
                raise ValueError(
                    "Class {} failed to provide an (args, kwargs) tuple and "
                    "provided instead: {}".format(typename, repr(result)))
        return args_kwargs_to_initdict(*result)

    return wrapper


# =============================================================================
# Function to make custom instance -> InitDict functions
# =============================================================================

def make_instance_to_initdict(attributes: List[str]) -> InstanceToDictFnType:
    def custom_instance_to_initdict(x: Instance) -> InitDict:
        kwargs = {}
        for a in attributes:
            kwargs[a] = getattr(x, a)
        return kwargs_to_initdict(kwargs)

    return custom_instance_to_initdict


# =============================================================================
# Describe how a Python class should be serialized to/from JSON
# =============================================================================

class JsonDescriptor(object):
    def __init__(self,
                 typename: str,
                 obj_to_dict_fn: InstanceToDictFnType,
                 dict_to_obj_fn: DictToInstanceFnType,
                 cls: ClassType,
                 default_factory: DefaultFactoryFnType = None) -> None:
        self._typename = typename
        self._obj_to_dict_fn = obj_to_dict_fn
        self._dict_to_obj_fn = dict_to_obj_fn
        self._cls = cls
        self._default_factory = default_factory

    def to_dict(self, obj: Instance) -> Dict:
        return self._obj_to_dict_fn(obj)

    def to_obj(self, d: Dict) -> Instance:
        # noinspection PyBroadException
        try:
            return self._dict_to_obj_fn(d, self._cls)
        except Exception as err:
            log.warning(
                "Failed to deserialize object of type {t}; exception was {e}; "
                "dict was {d}; will use default factory instead".format(
                    t=self._typename, e=repr(err), d=repr(d)))
            if self._default_factory:
                return self._default_factory()
            else:
                return None

    def __repr__(self):
        return (
            "<{qualname}("
            "typename={typename}, "
            "obj_to_dict_fn={obj_to_dict_fn}, "
            "dict_to_obj_fn={dict_to_obj_fn}, "
            "cls={cls}, "
            "default_factory={default_factory}"
            ") at {addr}>".format(
                qualname=self.__class__.__qualname__,
                typename=repr(self._typename),
                obj_to_dict_fn=repr(self._obj_to_dict_fn),
                dict_to_obj_fn=repr(self._dict_to_obj_fn),
                cls=repr(self._cls),
                default_factory=repr(self._default_factory),
                addr=hex(id(self)),
            )
        )


# =============================================================================
# Maintain a record of how several classes should be serialized
# =============================================================================

TYPE_MAP = {}  # type: Dict[str, JsonDescriptor]


def register_class_for_json(
        cls: ClassType,
        method: str = METHOD_SIMPLE,
        obj_to_dict_fn: InstanceToDictFnType = None,
        dict_to_obj_fn: DictToInstanceFnType = initdict_to_instance,
        default_factory: DefaultFactoryFnType = None) -> None:
    """
    Registers the class cls for JSON serialization.
    - If both obj_to_dict_fn and dict_to_obj_fn are registered, the framework
      uses these to convert instances of the class to/from Python dictionaries,
      which are in turn serialized to JSON.
    - Otherwise:

        if method == 'simple':
            # ... uses simple_to_dict and simple_from_dict (q.v.)

        if method == 'strip_underscore':
            # ... uses strip_underscore_to_dict and simple_from_dict (q.v.)
    """
    typename = cls.__qualname__  # preferable to __name__
    # ... __name__ looks like "Thing" and is ambiguous
    # ... __qualname__ looks like "my.module.Thing" and is not
    if obj_to_dict_fn and dict_to_obj_fn:
        descriptor = JsonDescriptor(
            typename=typename,
            obj_to_dict_fn=obj_to_dict_fn,
            dict_to_obj_fn=dict_to_obj_fn,
            cls=cls,
            default_factory=default_factory)
    elif method == METHOD_SIMPLE:
        descriptor = JsonDescriptor(
            typename=typename,
            obj_to_dict_fn=instance_to_initdict_simple,
            dict_to_obj_fn=initdict_to_instance,
            cls=cls,
            default_factory=default_factory)
    elif method == METHOD_STRIP_UNDERSCORE:
        descriptor = JsonDescriptor(
            typename=typename,
            obj_to_dict_fn=instance_to_initdict_stripping_underscores,
            dict_to_obj_fn=initdict_to_instance,
            cls=cls,
            default_factory=default_factory)
    else:
        raise ValueError("Unknown method, and functions not fully specified")
    global TYPE_MAP
    TYPE_MAP[typename] = descriptor


def register_for_json(*args, **kwargs) -> Any:
    """
    Class decorator to register classes with our JSON system.

    - If method is 'provides_init_args_kwargs', the class provides a
      function

        def init_args_kwargs(self) -> Tuple[List[Any], Dict[str, Any]]

      that returns an (args, kwargs) tuple, suitable for passing to its
      __init__() function as __init__(*args, **kwargs).

    - If method is 'provides_init_kwargs', the class provides a function

        def init_kwargs(self) -> Dict

      that returns a dictionary kwargs suitable for passing to its __init__()
      function as __init__(**kwargs).

    - Otherwise, the method argument is as for register_class_for_json().

    Usage looks like:

    @register_for_json(method=METHOD_STRIP_UNDERSCORE)
    class TableId(object):
        def __init__(self, db: str = '', schema: str = '',
                     table: str = '') -> None:
            self._db = db
            self._schema = schema
            self._table = table

    """
    if DEBUG:
        print("register_for_json: args = {}".format(repr(args)))
        print("register_for_json: kwargs = {}".format(repr(kwargs)))

    # http://stackoverflow.com/questions/653368/how-to-create-a-python-decorator-that-can-be-used-either-with-or-without-paramet  # noqa
    # In brief,
    #   @decorator
    #   x
    #
    # means
    #   x = decorator(x)
    #
    # so
    #   @decorator(args)
    #   x
    #
    # means
    #   x = decorator(args)(x)

    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        if DEBUG:
            print("... called as @register_for_json")
        # called as @decorator
        # ... the single argument is the class itself, e.g. Thing in:
        #   @decorator
        #   class Thing(object):
        #       # ...
        # ... e.g.:
        #   args = (<class '__main__.unit_tests.<locals>.SimpleThing'>,)
        #   kwargs = {}
        cls = args[0]  # type: ClassType
        register_class_for_json(cls, method=METHOD_SIMPLE)
        return cls

    # Otherwise:
    if DEBUG:
        print("... called as @register_for_json(*args, **kwargs)")
    # called as @decorator(*args, **kwargs)
    # ... e.g.:
    #   args = ()
    #   kwargs = {'method': 'provides_to_init_args_kwargs_dict'}
    method = kwargs.pop('method', METHOD_SIMPLE)  # type: str
    obj_to_dict_fn = kwargs.pop('obj_to_dict_fn', None)  # type: InstanceToDictFnType  # noqa
    dict_to_obj_fn = kwargs.pop('dict_to_obj_fn', initdict_to_instance)  # type: DictToInstanceFnType  # noqa
    default_factory = kwargs.pop('default_factory', None)  # type: DefaultFactoryFnType  # noqa
    check_result = kwargs.pop('check_results', True)  # type: bool

    def register_json_class(cls_: ClassType) -> ClassType:
        odf = obj_to_dict_fn
        dof = dict_to_obj_fn
        if method == METHOD_PROVIDES_INIT_ARGS_KWARGS:
            if hasattr(cls_, INIT_ARGS_KWARGS_FN_NAME):
                odf = wrap_args_kwargs_to_initdict(
                    getattr(cls_, INIT_ARGS_KWARGS_FN_NAME),
                    typename=cls_.__qualname__,
                    check_result=check_result
                )
            else:
                raise ValueError(
                    "Class type {} does not provide function {}".format(
                        cls_, INIT_ARGS_KWARGS_FN_NAME))
        elif method == METHOD_PROVIDES_INIT_KWARGS:
            if hasattr(cls_, INIT_KWARGS_FN_NAME):
                odf = wrap_kwargs_to_initdict(
                    getattr(cls_, INIT_KWARGS_FN_NAME),
                    typename=cls_.__qualname__,
                    check_result=check_result
                )
            else:
                raise ValueError(
                    "Class type {} does not provide function {}".format(
                        cls_, INIT_KWARGS_FN_NAME))
        elif method == METHOD_NO_ARGS:
            odf = obj_with_no_args_to_init_dict
        register_class_for_json(cls_,
                                method=method,
                                obj_to_dict_fn=odf,
                                dict_to_obj_fn=dof,
                                default_factory=default_factory)
        return cls_

    return register_json_class


def dump_map(file: TextIO = sys.stdout) -> None:
    pp = pprint.PrettyPrinter(indent=4, stream=file)
    print("Type map: ", file=file)
    pp.pprint(TYPE_MAP)


# =============================================================================
# Hooks to implement the JSON encoding/decoding
# =============================================================================

class JsonClassEncoder(json.JSONEncoder):
    def default(self, obj: Instance) -> Any:
        typename = type(obj).__qualname__  # preferable to __name__, as above
        if typename in TYPE_MAP:
            descriptor = TYPE_MAP[typename]
            d = descriptor.to_dict(obj)
            if TYPE_LABEL in d:
                raise ValueError("Class already has attribute: " + TYPE_LABEL)
            d[TYPE_LABEL] = typename
            if DEBUG:
                log.debug("Serializing {} -> {}".format(repr(obj), repr(d)))
            return d
        # Otherwise, nothing that we know about:
        return super().default(obj)


def json_class_decoder_hook(d: Dict) -> Any:
    if TYPE_LABEL in d:
        typename = d.get(TYPE_LABEL)
        if typename in TYPE_MAP:
            if DEBUG:
                log.debug("Deserializing: {}".format(repr(d)))
            d.pop(TYPE_LABEL)
            descriptor = TYPE_MAP[typename]
            obj = descriptor.to_obj(d)
            if DEBUG:
                log.debug("... to: {}".format(repr(obj)))
            return obj
    return d


# =============================================================================
# Functions for end users
# =============================================================================

def json_encode(obj: Instance) -> str:
    return json.dumps(obj, cls=JsonClassEncoder)


def json_decode(s: str) -> Any:
    try:
        return json.JSONDecoder(object_hook=json_class_decoder_hook).decode(s)
    except json.JSONDecodeError:
        log.warning("Failed to decode JSON (returning None): {}".format(
            repr(s)))
        return None


# =============================================================================
# Django field
# - To use a class with this, the class must be registered with
#   register_class_for_json() above. Register the class immediately after
#   defining it.
# =============================================================================

class JsonClassField(TextField):
    # https://docs.djangoproject.com/en/1.10/howto/custom-model-fields/
    description = "Python objects serialized into JSON"

    # No need to implement __init__()
    # No need to implement deconstruct()
    # No need to implement db_type()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def from_db_value(self, value, expression, connection, context):
        """
        "Called in all circumstances when the data is loaded from the
        database, including in aggregates and values() calls."
        """
        if value is None:
            return value
        return json_decode(value)

    def to_python(self, value):
        """
        "Called during deserialization and during the clean() method used
        from forms.... [s]hould deal gracefully with... (*) an instance of
        the correct type; (*) a string; (*) None (if the field allows
        null=True)."

        "For to_python(), if anything goes wrong during value conversion, you
        should raise a ValidationError exception."
        """
        if value is None:
            return value
        if not isinstance(value, str):
            return value
        try:
            return json_decode(value)
        except Exception as err:
            raise ValidationError(repr(err))

    def get_prep_value(self, value):
        """
        Converse of to_python(). Converts Python objects back to query
        values.
        """
        return json_encode(value)


# =============================================================================
# Implement JSON translation for common types
# =============================================================================

register_class_for_json(
    cls=datetime.date,
    obj_to_dict_fn=make_instance_to_initdict(['year', 'month', 'day'])
)
register_class_for_json(
    cls=datetime.datetime,
    obj_to_dict_fn=make_instance_to_initdict([
        'year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond'
    ])
)
# Note in passing: the repr() of datetime.date and datetime.datetime look like
# 'datetime.date(...)' only because their repr() function explicitly does
# 'datetime.' + self.__class__.__name__; there's no way, it seems, to get
# that or __qualname__ to add the prefix automatically.


# =============================================================================
# Testing
# =============================================================================

def simple_eq(one: Instance, two: Instance, attrs: List[str]) -> bool:
    return all(getattr(one, a) == getattr(two, a) for a in attrs)


def unit_tests():

    class BaseTestClass(object):
        def __repr__(self) -> str:
            return auto_repr(self, with_addr=True)

        def __str__(self) -> str:
            return repr(self)

    @register_for_json
    class SimpleThing(BaseTestClass):
        def __init__(self, a, b, c, d: datetime.datetime = None):
            self.a = a
            self.b = b
            self.c = c
            self.d = d or datetime.datetime.now()

        def __eq__(self, other: 'SimpleThing') -> bool:
            return simple_eq(self, other, ['a', 'b', 'c', 'd'])

    # If you comment out the decorator for this derived class, serialization
    # will fail, and that is a good thing (derived classes shouldn't be
    # serialized on a "have a try" basis).
    @register_for_json
    class DerivedThing(BaseTestClass):
        def __init__(self, a, b, c, d: datetime.datetime = None, e: int = 5,
                     f: datetime.date = None):
            self.a = a
            self.b = b
            self.c = c
            self.d = d or datetime.datetime.now()
            self.e = e
            self.f = f or datetime.date.today()

        def __eq__(self, other: 'SimpleThing') -> bool:
            return simple_eq(self, other, ['a', 'b', 'c', 'd', 'e'])

    @register_for_json(method=METHOD_STRIP_UNDERSCORE)
    class UnderscoreThing(BaseTestClass):
        def __init__(self, a, b, c):
            self._a = a
            self._b = b
            self._c = c

        # noinspection PyProtectedMember
        def __eq__(self, other: 'UnderscoreThing') -> bool:
            return simple_eq(self, other, ['_a', '_b', '_c'])

    @register_for_json(method=METHOD_PROVIDES_INIT_ARGS_KWARGS)
    class InitDictThing(BaseTestClass):
        def __init__(self, a, b, c):
            self.p = a
            self.q = b
            self.r = c

        def __eq__(self, other: 'InitDictThing') -> bool:
            return simple_eq(self, other, ['p', 'q', 'r'])

        def init_args_kwargs(self) -> ArgsKwargsTuple:
            args = []
            kwargs = {'a': self.p, 'b': self.q, 'c': self.r}
            return args, kwargs

    @register_for_json(method=METHOD_PROVIDES_INIT_KWARGS)
    class KwargsDictThing(BaseTestClass):
        def __init__(self, a, b, c):
            self.p = a
            self.q = b
            self.r = c

        def __eq__(self, other):
            return simple_eq(self, other, ['p', 'q', 'r'])

        def init_kwargs(self) -> KwargsDict:
            return {'a': self.p, 'b': self.q, 'c': self.r}

    def check_json(start: Any) -> None:
        print(repr(start))
        encoded = json_encode(start)
        print("-> JSON: " + repr(encoded))
        resurrected = json_decode(encoded)
        print("-> resurrected: " + repr(resurrected))
        assert resurrected == start
        print("... OK")
        print()

    check_json(SimpleThing(1, 2, 3))
    check_json(DerivedThing(1, 2, 3, e=6))
    check_json(UnderscoreThing(1, 2, 3))
    check_json(InitDictThing(1, 2, 3))
    check_json(KwargsDictThing(1, 2, 3))

    dump_map()

    print("\nAll OK.\n")


if __name__ == '__main__':
    unit_tests()
